# -*- coding: utf-8 -*-

from application.model.customexception import NoUpdateFound, DisabledFunctionality
from application.repository.pos import CatalogUpdaterPOSRepository
from application.service import UpdaterService
from application.util import LoggerUtil
from helper import remove_accents

logger = LoggerUtil.get_logger_name(__name__)


class GetCatalogVersionToApplyInteractor:
    def __init__(self, pos_repo, updater_service, enabled):
        # type: (CatalogUpdaterPOSRepository, UpdaterService, bool) -> None
        self.pos_repo = pos_repo
        self.updater_service = updater_service
        self.enabled = enabled

    def get_catalog_version_to_apply(self, need_backoffice_json):
        if not self.enabled:
            raise DisabledFunctionality()

        downloaded_version_name = self.updater_service.download_campaign_pending_version(need_backoffice_json)
        if not downloaded_version_name:
            logger.info("No updates available")
            raise NoUpdateFound()

        downloaded_version_name = remove_accents(downloaded_version_name)
        logger.info("Available update version {} to apply".format(downloaded_version_name))
        return downloaded_version_name



# vat.py - functions for handling Swiss VAT numbers
# coding: utf-8
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VAT, MWST, TVA, IVA, TPV (Mehrwertsteuernummer, the Swiss VAT number).

The Swiss VAT number is based on the UID but is followed by either "MWST"
(Mehrwertsteuer, the German abbreviation for VAT), "TVA" (Taxe sur la valeur
ajoutée in French), "IVA" (Imposta sul valore aggiunto in Italian) or "TPV"
(Taglia sin la plivalur in Romanian).

This module only supports the "new" format that was introduced in 2011 which
completely replaced the "old" 6-digit format in 2014.

More information:

* https://www.ch.ch/en/value-added-tax-number-und-business-identification-number/
* https://www.uid.admin.ch/

>>> validate('CHE-107.787.577 IVA')
'CHE107787577IVA'
>>> validate('CHE-107.787.578 IVA')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('CHE107787577IVA')
'CHE-107.787.577 IVA'
"""

from stdnum.ch import uid
from stdnum.exceptions import *


def compact(number):
    """Convert the number to the minimal representation. This strips
    surrounding whitespace and separators."""
    return uid.compact(number)


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) not in (15, 16):
        raise InvalidLength()
    uid.validate(number[:12])
    if number[12:] not in ('MWST', 'TVA', 'IVA', 'TPV'):
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return uid.format(number[:12]) + ' ' + number[12:]



import base64
import json
from xml.etree import cElementTree as eTree

import sysactions
from helper import remove_accents
from modelcommon.printer import SeparatorTag
from msgbus import (
    FM_PARAM,
    TK_PRN_PRINT,
    TK_SYS_ACK,
)
from sysactions import (
    get_storewide_config,
    get_model,
    get_current_order,
    show_confirmation,
    get_podtype,
)
from typing import (
    Dict,
    Optional,
    List,
)

from constants import (
    TXT_BOLD_ON,
    CTL_LF,
    TXT_BOLD_OFF,
    TXT_ALIGN_CT,
    TXT_FONT_A,
    TXT_2WIDTH,
    TXT_2HEIGHT,
    HW_INIT,
    TXT_NORMAL,
    TXT_ALIGN_LT,
)

from utilfunctions import (
    get_device_service,
    is_table_service,
)


print_merchant_receipt = None
print_customer_receipt = None
ask_print_customer_receipt = None


class PrintReceipt(object):
    def __init__(self, mb_context, logger):
        global print_merchant_receipt
        global print_customer_receipt
        global ask_print_customer_receipt

        self.mb_context = mb_context
        self.logger = logger

        if print_merchant_receipt is None:
            print_merchant_receipt = self._get_store_config("Store.PrintMerchantReceipt", "False")
        self.print_merchant_receipt = print_merchant_receipt

        if print_customer_receipt is None:
            print_customer_receipt = self._get_store_config("Store.PrintCustomerReceipt", "False")
        self.print_customer_receipt = print_customer_receipt

        if ask_print_customer_receipt is None:
            ask_print_customer_receipt = self._get_store_config("Store.AskPrintCustomerReceipt", "True")
        self.ask_print_customer_receipt = ask_print_customer_receipt

    @staticmethod
    def _get_store_config(
        config,         # type: str
        default_value,  # type: str
    ):
        # type: (...) -> bool

        return get_storewide_config(config, defval=default_value).lower() == "true"

    def print_tef_receipts(
        self,
        pos_id,                     # type: str
        fiscalization_type="",      # type: Optional[str]
        order_id=None,              # type: Optional[str]
    ):
        # type: (...) -> None

        try:
            model = get_model(pos_id)
            if not order_id:
                order_id = get_current_order(model).get("orderId")

            order_taker = sysactions.get_posot(model)
            order_pict = order_taker.orderPicture(orderid=order_id)
            order = eTree.XML(order_pict).find("Order")
            tenders = order.findall("TenderHistory/Tender")
            print_customer_receipt_response = None
            sector_name = None
            sector = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("SECTOR"))
            if sector is not None:
                sector_name = sector.get("value", None)

            total_electronic_tenders = 1
            for tender in tenders:
                if int(tender.get("eletronicType")) not in (0, 9):

                    if float(tender.get("tenderAmount")) <= 0:
                        continue

                    tender_detail = json.loads(tender.get("tenderDetail"))
                    print_customer_receipt_response = self.print_tef_tender_receipt(
                        pos_id=pos_id,
                        tender_detail=tender_detail,
                        fiscalization_type=fiscalization_type,
                        print_customer_receipt=print_customer_receipt_response,
                        sector=sector_name,
                        total_electronic_tenders=total_electronic_tenders,
                    )
                    total_electronic_tenders += 1
        except Exception:  # noqa
            self.logger.exception("Error printing TEF receipts")

    def print_tef_tender_receipt(
        self,
        pos_id,                         # type: str
        tender_detail,                  # type: Dict
        fiscalization_type="",          # type: Optional[str]
        print_customer_receipt=None,    # type: Optional[bool]
        sector=None,                    # type: str
        total_electronic_tenders=1,     # type: Optional[int]
    ):
        # type: (...) -> Optional[bool]

        has_merchant_receipt = "ReceiptMerchant" in tender_detail and tender_detail["ReceiptMerchant"] not in ['', None]
        has_customer_receipt = "ReceiptCustomer" in tender_detail and tender_detail["ReceiptCustomer"] not in ['', None]

        if has_merchant_receipt or has_customer_receipt:
            try:
                model = get_model(pos_id)
                pod_type = get_podtype(model)
                is_table = is_table_service(model=model)
                current_printer = get_device_service(
                    pos_id=pos_id,
                    device_type="printer",
                    sector=sector,
                )

                if has_merchant_receipt and self.print_merchant_receipt and pod_type != "TT":
                    self.print_tender_receipt(
                        pos_id=pos_id,
                        current_printer=current_printer,
                        receipt_data=tender_detail,
                        receipt_type="ReceiptMerchant",
                        fiscalization_type=fiscalization_type,
                        is_table=is_table,
                        total_electronic_tenders=total_electronic_tenders,
                    )

                if has_customer_receipt:
                    ignore_confirmation_message = False
                    if self.ask_print_customer_receipt:
                        if print_customer_receipt is None or total_electronic_tenders == 1:
                            print_customer_receipt = show_confirmation(pos_id, "$WANT_TO_PRINT_CUSTOMER_RECEIPT")
                            ignore_confirmation_message = print_customer_receipt
                    else:
                        print_customer_receipt = self.print_customer_receipt

                    if print_customer_receipt:
                        self.print_tender_receipt(
                            pos_id=pos_id,
                            current_printer=current_printer,
                            receipt_data=tender_detail,
                            receipt_type="ReceiptCustomer",
                            fiscalization_type=fiscalization_type,
                            is_table=is_table,
                            total_electronic_tenders=total_electronic_tenders,
                            ignore_confirmation_message=ignore_confirmation_message,
                        )
            except Exception:  # noqa
                self.logger.exception("Error printing TEF receipt for tender")

            return print_customer_receipt

    def print_tender_receipt(
        self,
        pos_id,                                 # type: str
        current_printer,                        # type: str
        receipt_data,                           # type: Dict | str
        receipt_type=None,                      # type: Optional[str]
        fiscalization_type="",                  # type: Optional[str]
        reprint_text=None,                      # type: Optional[str]
        is_table=False,                         # type: Optional[bool]
        total_electronic_tenders=1,             # type: Optional[int]
        ignore_confirmation_message=False,      # type: Optional[bool]
    ):
        # type: (...) -> bool

        if fiscalization_type.startswith("APOS") and not is_table and not ignore_confirmation_message:
            self._show_confirmation_message(
                pos_id=pos_id,
                receipt_type=receipt_type,
                total_electronic_tenders=total_electronic_tenders,
            )

        receipt_data = receipt_data if receipt_type is None else receipt_data[receipt_type]
        receipt = self._decode_receipt(
            receipt_data=receipt_data,
            fiscalization_type=fiscalization_type,
            reprint_text=reprint_text,
        )

        resp = self.mb_context.MB_EasySendMessage(
            dest_name=current_printer,
            token=TK_PRN_PRINT,
            format=FM_PARAM,
            data=receipt,
            timeout=30000000,
        )

        return resp.token == TK_SYS_ACK

    def reprint_receipt(
        self,
        pos_id,                     # type: str
        receipts,                   # type: List[str]
        model=None,                 # type: Optional[eTree]
        fiscalization_type="",      # type: str
        print_reprint_text=True,    # type: bool
    ):

        # type: (...) -> bool

        if model is None:
            model = sysactions.get_model(posid=pos_id)

        current_printer = get_device_service(
            pos_id=pos_id,
            device_type="printer",
        )

        reprint_text = ""
        if print_reprint_text:
            reprint_text = sysactions.translate_message(
                model=model,
                message_key="REPRINT_TEXT_MESSAGE",
            )

        reprinted = True
        for receipt in receipts:
            reprinted &= self.print_tender_receipt(
                pos_id=pos_id,
                current_printer=current_printer,
                receipt_data=receipt,
                fiscalization_type=fiscalization_type,
                reprint_text=reprint_text,
            )

        return reprinted

    @staticmethod
    def _show_confirmation_message(
        pos_id,                         # type: str
        receipt_type,                   # type: str
        total_electronic_tenders,       # type: int
    ):
        # type: (...) -> None

        if total_electronic_tenders > 1 or receipt_type == "ReceiptMerchant":
            customer_message = "$CONFIRMATION_MESSAGE_PRINT_RECEIPT_CUSTOMER"
            merchant_message = "$CONFIRMATION_MESSAGE_PRINT_RECEIPT_MERCHANT"
            merchant_next_message = "$CONFIRMATION_MESSAGE_NEXT_PRINT_RECEIPT_MERCHANT"

            is_next_message_merchant = total_electronic_tenders > 1 and receipt_type == "ReceiptMerchant"
            merchant_message = merchant_next_message if is_next_message_merchant else merchant_message

            message = customer_message if receipt_type == "ReceiptCustomer" else merchant_message

            show_confirmation(
                posid=pos_id,
                message=message,
                buttons="$OK",
            )

    def _decode_receipt(
        self,
        receipt_data,           # type: str
        fiscalization_type,     # type: str
        reprint_text,           # type: str
    ):
        # type: (...) -> str

        receipt_data = base64.b64decode(receipt_data)

        if reprint_text not in ['', None]:
            reprint_text = self._get_reprint_text(
                reprint_text=reprint_text,
                fiscalization_type=fiscalization_type,
            )

            reprint_text = remove_accents(
                text=reprint_text,
            )

            receipt_data = "{}{}{}".format(
                receipt_data,
                CTL_LF,
                reprint_text,
            )

        append_cut_command = not fiscalization_type or not fiscalization_type.startswith("APOS")
        cut_tag = b"{}SpaceCutReceiptCard{}"
        space_cut_receipt_card = cut_tag.format(SeparatorTag.start, SeparatorTag.end) if append_cut_command else ""
        receipt = "{}{}{}".format(
            TXT_ALIGN_CT,
            receipt_data,
            space_cut_receipt_card,
        )

        return receipt

    @staticmethod
    def _get_reprint_text(
        reprint_text,            # type: str
        fiscalization_type,      # type: str
    ):
        # type: (...) -> str

        if fiscalization_type.startswith("APOS"):
            max_width = 38
            pre_space = int((max_width - len(reprint_text)) / 2)
            post_space = max_width - len(reprint_text) - pre_space
            return (" " * pre_space) + reprint_text + (" " * post_space)

        rollback_printer_default_font = "{}{}{}{}{}".format(
            HW_INIT,
            TXT_NORMAL,
            TXT_BOLD_OFF,
            TXT_FONT_A,
            TXT_ALIGN_LT,
        )

        return b"{}{}{}{}{}{}{}".format(
            CTL_LF,
            TXT_2HEIGHT,
            TXT_2WIDTH,
            TXT_BOLD_ON,
            reprint_text,
            CTL_LF,
            rollback_printer_default_font,
        )



from typing import (
    Union,
    Optional,
)
from xml.etree import cElementTree as eTree

from sysactions import (
    show_print_preview,
    show_messagebox,
)

from actions.printer import send_json_data_print
from json_report_to_html import JSONReportParserToHtml


def print_preview_report(
    pos_id,               # type: Union[str, int]
    report_data,          # type: str
    preview=True,         # type: Optional[bool]
    order=None,           # type: Optional[eTree.Element]
    allow_printing=True,  # type: Optional[bool]
):
    # type: (...) -> bool

    if report_data is not None:
        if preview:
            report_preview = JSONReportParserToHtml().parse(report=report_data)
            result = show_print_preview(
                posid=pos_id,
                text_data=report_preview.encode("utf-8"),
                title="$PRINT_PREVIEW",
                buttons="$PRINT|$CLOSE" if allow_printing else "$CLOSE",
            )
            if result in (1, None) or not allow_printing:
                return False

        send_json_data_print(
            pos_id=pos_id,
            print_data=report_data,
            order=order,
        )
    elif report_data in [None, "", "{}", "[]"]:
        show_messagebox(
            posid=pos_id,
            message="$NO_REPORT_DATA",
            title="$ERROR",
        )



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives import ciphers
from cryptography.hazmat.primitives.ciphers import modes

@typing.overload
def create_encryption_ctx(
    algorithm: ciphers.CipherAlgorithm, mode: modes.ModeWithAuthenticationTag
) -> ciphers.AEADEncryptionContext: ...
@typing.overload
def create_encryption_ctx(
    algorithm: ciphers.CipherAlgorithm, mode: modes.Mode
) -> ciphers.CipherContext: ...
@typing.overload
def create_decryption_ctx(
    algorithm: ciphers.CipherAlgorithm, mode: modes.ModeWithAuthenticationTag
) -> ciphers.AEADDecryptionContext: ...
@typing.overload
def create_decryption_ctx(
    algorithm: ciphers.CipherAlgorithm, mode: modes.Mode
) -> ciphers.CipherContext: ...
def cipher_supported(
    algorithm: ciphers.CipherAlgorithm, mode: modes.Mode
) -> bool: ...
def _advance(
    ctx: ciphers.AEADEncryptionContext | ciphers.AEADDecryptionContext, n: int
) -> None: ...
def _advance_aad(
    ctx: ciphers.AEADEncryptionContext | ciphers.AEADDecryptionContext, n: int
) -> None: ...

class CipherContext: ...
class AEADEncryptionContext: ...
class AEADDecryptionContext: ...



from ._TableError import TableError
from typing import Dict


class TabToJoinInInvalidStatus(TableError):

    def __init__(
        self,       # type: TabToJoinInInvalidStatus
        state,      # type: str
        tab_name,   # type: str
    ):
        # type: (...) -> None

        message = "$TAB_TO_JOIN_IN_INVALID_STATUS|{}|{}".format(tab_name, state)
        super(TabToJoinInInvalidStatus, self).__init__(message)

        self.tab_name = tab_name
        self.state = state

    def get_extra_data(
        self,   # type: TabToJoinInInvalidStatus
    ):
        # type: (...) -> Dict[str, str]

        return {
            "tabName": self.tab_name,
            "state": self.state,
        }



# A binary morphology add-on for the Python Imaging Library
#
# History:
#   2014-06-04 Initial version.
#
# Copyright (c) 2014 Dov Grobgeld <dov.grobgeld@gmail.com>

from __future__ import print_function

import re

from . import Image, _imagingmorph

LUT_SIZE = 1 << 9

# fmt: off
ROTATION_MATRIX = [
    6, 3, 0,
    7, 4, 1,
    8, 5, 2,
]
MIRROR_MATRIX = [
    2, 1, 0,
    5, 4, 3,
    8, 7, 6,
]
# fmt: on


class LutBuilder(object):
    """A class for building a MorphLut from a descriptive language

      The input patterns is a list of a strings sequences like these::

          4:(...
             .1.
             111)->1

      (whitespaces including linebreaks are ignored). The option 4
      describes a series of symmetry operations (in this case a
      4-rotation), the pattern is described by:

      - . or X - Ignore
      - 1 - Pixel is on
      - 0 - Pixel is off

      The result of the operation is described after "->" string.

      The default is to return the current pixel value, which is
      returned if no other match is found.

      Operations:

      - 4 - 4 way rotation
      - N - Negate
      - 1 - Dummy op for no other operation (an op must always be given)
      - M - Mirroring

      Example::

          lb = LutBuilder(patterns = ["4:(... .1. 111)->1"])
          lut = lb.build_lut()

    """

    def __init__(self, patterns=None, op_name=None):
        if patterns is not None:
            self.patterns = patterns
        else:
            self.patterns = []
        self.lut = None
        if op_name is not None:
            known_patterns = {
                "corner": ["1:(... ... ...)->0", "4:(00. 01. ...)->1"],
                "dilation4": ["4:(... .0. .1.)->1"],
                "dilation8": ["4:(... .0. .1.)->1", "4:(... .0. ..1)->1"],
                "erosion4": ["4:(... .1. .0.)->0"],
                "erosion8": ["4:(... .1. .0.)->0", "4:(... .1. ..0)->0"],
                "edge": [
                    "1:(... ... ...)->0",
                    "4:(.0. .1. ...)->1",
                    "4:(01. .1. ...)->1",
                ],
            }
            if op_name not in known_patterns:
                raise Exception("Unknown pattern " + op_name + "!")

            self.patterns = known_patterns[op_name]

    def add_patterns(self, patterns):
        self.patterns += patterns

    def build_default_lut(self):
        symbols = [0, 1]
        m = 1 << 4  # pos of current pixel
        self.lut = bytearray(symbols[(i & m) > 0] for i in range(LUT_SIZE))

    def get_lut(self):
        return self.lut

    def _string_permute(self, pattern, permutation):
        """string_permute takes a pattern and a permutation and returns the
        string permuted according to the permutation list.
        """
        assert len(permutation) == 9
        return "".join(pattern[p] for p in permutation)

    def _pattern_permute(self, basic_pattern, options, basic_result):
        """pattern_permute takes a basic pattern and its result and clones
        the pattern according to the modifications described in the $options
        parameter. It returns a list of all cloned patterns."""
        patterns = [(basic_pattern, basic_result)]

        # rotations
        if "4" in options:
            res = patterns[-1][1]
            for i in range(4):
                patterns.append(
                    (self._string_permute(patterns[-1][0], ROTATION_MATRIX), res)
                )
        # mirror
        if "M" in options:
            n = len(patterns)
            for pattern, res in patterns[0:n]:
                patterns.append((self._string_permute(pattern, MIRROR_MATRIX), res))

        # negate
        if "N" in options:
            n = len(patterns)
            for pattern, res in patterns[0:n]:
                # Swap 0 and 1
                pattern = pattern.replace("0", "Z").replace("1", "0").replace("Z", "1")
                res = 1 - int(res)
                patterns.append((pattern, res))

        return patterns

    def build_lut(self):
        """Compile all patterns into a morphology lut.

        TBD :Build based on (file) morphlut:modify_lut
        """
        self.build_default_lut()
        patterns = []

        # Parse and create symmetries of the patterns strings
        for p in self.patterns:
            m = re.search(r"(\w*):?\s*\((.+?)\)\s*->\s*(\d)", p.replace("\n", ""))
            if not m:
                raise Exception('Syntax error in pattern "' + p + '"')
            options = m.group(1)
            pattern = m.group(2)
            result = int(m.group(3))

            # Get rid of spaces
            pattern = pattern.replace(" ", "").replace("\n", "")

            patterns += self._pattern_permute(pattern, options, result)

        # compile the patterns into regular expressions for speed
        for i, pattern in enumerate(patterns):
            p = pattern[0].replace(".", "X").replace("X", "[01]")
            p = re.compile(p)
            patterns[i] = (p, pattern[1])

        # Step through table and find patterns that match.
        # Note that all the patterns are searched. The last one
        # caught overrides
        for i in range(LUT_SIZE):
            # Build the bit pattern
            bitpattern = bin(i)[2:]
            bitpattern = ("0" * (9 - len(bitpattern)) + bitpattern)[::-1]

            for p, r in patterns:
                if p.match(bitpattern):
                    self.lut[i] = [0, 1][r]

        return self.lut


class MorphOp(object):
    """A class for binary morphological operators"""

    def __init__(self, lut=None, op_name=None, patterns=None):
        """Create a binary morphological operator"""
        self.lut = lut
        if op_name is not None:
            self.lut = LutBuilder(op_name=op_name).build_lut()
        elif patterns is not None:
            self.lut = LutBuilder(patterns=patterns).build_lut()

    def apply(self, image):
        """Run a single morphological operation on an image

        Returns a tuple of the number of changed pixels and the
        morphed image"""
        if self.lut is None:
            raise Exception("No operator loaded")

        if image.mode != "L":
            raise Exception("Image must be binary, meaning it must use mode L")
        outimage = Image.new(image.mode, image.size, None)
        count = _imagingmorph.apply(bytes(self.lut), image.im.id, outimage.im.id)
        return count, outimage

    def match(self, image):
        """Get a list of coordinates matching the morphological operation on
        an image.

        Returns a list of tuples of (x,y) coordinates
        of all matching pixels. See :ref:`coordinate-system`."""
        if self.lut is None:
            raise Exception("No operator loaded")

        if image.mode != "L":
            raise Exception("Image must be binary, meaning it must use mode L")
        return _imagingmorph.match(bytes(self.lut), image.im.id)

    def get_on_pixels(self, image):
        """Get a list of all turned on pixels in a binary image

        Returns a list of tuples of (x,y) coordinates
        of all matching pixels. See :ref:`coordinate-system`."""

        if image.mode != "L":
            raise Exception("Image must be binary, meaning it must use mode L")
        return _imagingmorph.get_on_pixels(image.im.id)

    def load_lut(self, filename):
        """Load an operator from an mrl file"""
        with open(filename, "rb") as f:
            self.lut = bytearray(f.read())

        if len(self.lut) != LUT_SIZE:
            self.lut = None
            raise Exception("Wrong size operator file!")

    def save_lut(self, filename):
        """Save an operator to an mrl file"""
        if self.lut is None:
            raise Exception("No operator loaded")
        with open(filename, "wb") as f:
            f.write(self.lut)

    def set_lut(self, lut):
        """Set the lut from an external source"""
        self.lut = lut



# -*- coding: utf-8 -*-

# noinspection SpellCheckingInspection
class CfeTender(object):
    def __init__(
        self,               # type: CfeTender
        layout_version,     # type: str
    ):
        # type: (...) -> None

        self.layout_version = layout_version

        self.tpag_by_version = {
            "0.07": {
                CfeTenderType.dinheiro,
                CfeTenderType.cheque,
                CfeTenderType.cartao_credito,
                CfeTenderType.cartao_debito,
                CfeTenderType.credito_loja,
                CfeTenderType.vale_alimentacao,
                CfeTenderType.vale_refeicao,
                CfeTenderType.vale_presente,
                CfeTenderType.vale_combustivel,
                CfeTenderType.outros
            },
            "0.08": {
                CfeTenderType.dinheiro,
                CfeTenderType.cheque,
                CfeTenderType.cartao_credito,
                CfeTenderType.cartao_debito,
                CfeTenderType.credito_loja,
                CfeTenderType.vale_alimentacao,
                CfeTenderType.vale_refeicao,
                CfeTenderType.vale_presente,
                CfeTenderType.vale_combustivel,
                CfeTenderType.duplicata_mercanti,
                CfeTenderType.boleto,
                CfeTenderType.deposito_bancario,
                CfeTenderType.pagamento_instantaneo_pix,
                CfeTenderType.transferencia_bancaria__ou__carteira_digital,
                CfeTenderType.cash_back,
                CfeTenderType.programa_fidelidade__ou__cashback__ou__credito_virtual,
                CfeTenderType.sem_pagamento,
                CfeTenderType.outros
            }
        }

    def get_tpag(
        self,       # type: CfeTender
        tpag,       # type: int
    ):
        # type: (...) -> int

        pix_estatico = 20
        if tpag == pix_estatico:
            tpag = CfeTenderType.pagamento_instantaneo_pix

        if self.layout_version not in self.tpag_by_version:
            return CfeTenderType.outros

        if tpag not in self.tpag_by_version[self.layout_version]:
            return CfeTenderType.outros

        return tpag


# noinspection SpellCheckingInspection
class CfeTenderType(object):
    dinheiro = 1
    cheque = 2
    cartao_credito = 3
    cartao_debito = 4
    credito_loja = 5
    vale_alimentacao = 10
    vale_refeicao = 11
    vale_presente = 12
    vale_combustivel = 13
    duplicata_mercanti = 14
    boleto = 15
    deposito_bancario = 16
    pagamento_instantaneo_pix = 17
    transferencia_bancaria__ou__carteira_digital = 18
    cash_back = 19
    programa_fidelidade__ou__cashback__ou__credito_virtual = 19
    sem_pagamento = 90
    outros = 99



from api.discount import get_discount_service
from typing import List

global_discount_justifications = None


def get_global_discount_justifications():
    # type: (...) -> List[str]

    global global_discount_justifications

    if global_discount_justifications is None:
        global_discount_justifications = get_discount_service().get_justification_config().justifications

    return global_discount_justifications



import json
from logging import Logger

from bustoken import TK_REMOTE_ORDER_GET_DELIVERY_BRANDS
from msgbus import (
    MBEasyContext,
    TK_SYS_ACK
)
from typing import (
    List,
    Optional,
)


class DeliveryService(object):
    def __init__(
        self,
        mb_context,  # type: MBEasyContext
        logger       # type: Logger
    ):
        # type: (...) -> None

        self.mb_context = mb_context
        self.logger = logger
        self.delivery_brands = None

    def get_delivery_brands(self):
        # type: (...) -> Optional[List]

        if self.delivery_brands is not None:
            return self.delivery_brands

        try:
            result = self.mb_context.MB_EasySendMessage(
                dest_name="RemoteOrder",
                token=TK_REMOTE_ORDER_GET_DELIVERY_BRANDS,
                data=""
            )
            if result.token == TK_SYS_ACK:
                self.delivery_brands = json.loads(result.data)
        except (Exception,):
            self.logger.exception("Error getting delivery brands")
            return None

        return self.delivery_brands or []



class MaxSaleItemsQtyException(Exception):

    def __init__(
        self,    # type: MaxSaleItemsQtyException
        qty,     # type: int
    ):
        # type: (...) -> None

        super(MaxSaleItemsQtyException, self).__init__()

        self.message = "$MAX_ITEMS_TO_FISCALIZE|{}".format(qty)



# -*- coding: utf-8 -*-
import sqlite3
from logging import Logger

from cache_functions import UsersInfoCache
from dbrepository import SynchronizedDbApiTemplate
from messagebus import TokenCreator, TokenPriority, MessageBus
from messagehandler import MessageHandlerBuilder
from messagehandlerbuilder.token import ListenedEvents
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from production.box import ViewBox
from production.interactor import DeviceStatusChangedInteractor
from production.manager import ProductionManager, PicklistCleanerManager
from production.processor import (
    OrderModifiedProcessor,
    GetOrderMaxPrepTimeProcessor,
    GetOrderXmlProcessor,
    OrderModifiedIgnoredTypesProcessor,
    PurgeOrdersProcessor,
    GetPrintersInfoProcessor,
    PickListOrdersProcessor,
    DeviceStatusChangedProcessor,
    GetCoursesProcessor,
)
from production.processor.command import (
    SetOrderStateProcessor,
    ToggleTagLineProcessor,
    UndoProcessor,
    RefreshViewProcessor,
)
from production.treebuilder import TreeBuilder
from production.treebuilder.boxfactory import ProductionBoxFactory
from production.treebuilder.viewfactory import ViewFactory
from production.util import ProductionConfig
from sqliterepository import (
    ProductionRepository,
    ProductRepository,
)
from typing import (
    List,
    Dict,
    Union,
    Optional,
)

production_group = "C"
TK_PROD_REFRESH_VIEW = TokenCreator.create_token(TokenPriority.low, production_group, "1")
TK_PROD_SET_ORDER_STATE = TokenCreator.create_token(TokenPriority.low, production_group, "2")
TK_PROD_GET_ORDER_MAX_PREP_TIME = TokenCreator.create_token(TokenPriority.low, production_group, "3")
TK_PROD_GET_ORDER_XML = TokenCreator.create_token(TokenPriority.low, production_group, "4")  # 12582916
TK_PROD_UNDO = TokenCreator.create_token(TokenPriority.low, production_group, "7")
TK_PROD_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.low, production_group, "8")

TK_PROD_PURGE = TokenCreator.create_token(TokenPriority.low, production_group, "103")

TK_PROD_GET_PRINTERS_INFO = TokenCreator.create_token(TokenPriority.low, production_group, "201")
TK_PROD_GET_COURSES = TokenCreator.create_token(TokenPriority.low, production_group, "202")

TK_PROD_ORDER_PICKLIST = TokenCreator.create_token(TokenPriority.low, production_group, "301")  # 12583681


class ProductionMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(
        self,                           # type: ProductionMessageHandlerBuilder
        message_bus,                    # type: MessageBus
        purge_interval,                 # type: int
        orders_life_time,               # type: int
        manager_log_level,              # type: str
        audit_logger_prod_states,       # type: List
        kds_manager,                    # type: str
        production_config,              # type: ProductionConfig
        loggers,                        # type: Dict[unicode, Logger]
        timeout_to_cleanup_devices,     # type: int
        product_connection_string,      # type: str
        production_connection_string,   # type: str
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.purge_interval = purge_interval
        self.orders_life_time = orders_life_time
        self.manager_log_level = manager_log_level
        self.audit_logger_prod_states = audit_logger_prod_states
        self.production_config = production_config
        self.boxes_config = production_config.boxes_config
        self.branches_config = production_config.branches_config
        self.views_config = production_config.views_config
        self.clean_picklist_database_service_config = production_config.picklist_database_manager_config
        self.kds_manager = kds_manager
        self.loggers = loggers
        self.timeout_to_cleanup_devices = timeout_to_cleanup_devices
        self.product_connection_string = product_connection_string
        self.production_connection_string = production_connection_string
        self.production_repository = None   # type: Optional[ProductionRepository]
        self.product_repository = None      # type: Optional[ProductionRepository]
        self.production_connection = type   # type: Optional[sqlite3.connect]
        self.product_connection = type      # type: Optional[sqlite3.connect]

        self.production_manager = None  # type: Union[None, ProductionManager]
        self.message_handler = None
        self.user_cache_thread = None
        self.clean_picklist_database_service = None

    def build_singletons(self):
        if self.manager_log_level in self.loggers:
            manager_logger = self.loggers[self.manager_log_level]
        elif "ERROR" in self.loggers:
            manager_logger = self.loggers["ERROR"]
        elif len(self.loggers) > 0:
            manager_logger = self.loggers[self.loggers.keys()[0]]
        else:
            manager_logger = None

        self.product_repository = self._build_product_repository()
        self.production_repository = self._build_production_repository()

        self.production_manager = ProductionManager(
            self.production_repository,
            self.purge_interval,
            self.orders_life_time,
            self.message_bus,
            self.audit_logger_prod_states,
            self.kds_manager,
            manager_logger
        )

        production_box_factory = ProductionBoxFactory(
            self.production_manager,
            self.production_manager,
            self.production_manager,
            self.product_repository,
            self.loggers
        )

        view_factory = ViewFactory(
            message_bus=self.message_bus,
            loggers=self.loggers,
            production_manager=self.production_manager,
            kds_manager=self.kds_manager,
            timeout_to_cleanup_devices=self.timeout_to_cleanup_devices,
        )
        tree_builder = TreeBuilder(production_box_factory, view_factory)
        trees = tree_builder.build_trees(
            self.boxes_config,
            self.branches_config,
            self.views_config
        )

        root_boxes = []
        view_boxes = []
        for tree in trees:
            root_boxes.append(tree.root)
            leaves = [t for t in tree.leaves if isinstance(t, ViewBox)]
            view_boxes.extend(leaves)

        self.production_manager.set_root_boxes(root_boxes)
        self.production_manager.set_views(view_boxes)
        self.production_manager.initialize()

        self.user_cache_thread = UsersInfoCache(self.message_bus, manager_logger)
        self.user_cache_thread.daemon = True
        self.user_cache_thread.start()

        self.clean_picklist_database_service = PicklistCleanerManager(
            production_repository=self.production_repository,
            config=self.clean_picklist_database_service_config,
        )
        self.clean_picklist_database_service.start()

        order_modified_processor = OrderModifiedProcessor(
            self.production_manager,
            self.product_repository,
            self.user_cache_thread,
        )

        device_status_interactor = DeviceStatusChangedInteractor(production_manager=self.production_manager)
        device_status_changed_processor = DeviceStatusChangedProcessor(
            device_status_interactor=device_status_interactor,
        )

        event_processors = {
            ListenedEvents.DEVICE_STATUS_CHANGED: device_status_changed_processor,
        }

        order_modified_ignored_types_processor = OrderModifiedIgnoredTypesProcessor()
        event_processors.update({
            ListenedEvents.ORDER_MODIFIED: order_modified_processor,
            ListenedEvents.ORDER_MODIFIED_DO_TENDER: order_modified_ignored_types_processor,
            ListenedEvents.ORDER_MODIFIED_DISCOUNT_APPLIED: order_modified_ignored_types_processor,
            ListenedEvents.ORDER_MODIFIED_DISCOUNT_REMOVED: order_modified_ignored_types_processor,
            ListenedEvents.ORDER_MODIFIED_CLEAR_TENDERS: order_modified_ignored_types_processor,
            ListenedEvents.ORDER_MODIFIED_RESET_CURRENT_ORDER: order_modified_ignored_types_processor,
        })

        max_prep_time_processor = GetOrderMaxPrepTimeProcessor(self.production_manager, self.product_repository)
        message_processors = {
            TK_PROD_SET_ORDER_STATE: SetOrderStateProcessor(view_boxes),
            TK_PROD_TOGGLE_TAG_LINE: ToggleTagLineProcessor(view_boxes, self.loggers),
            TK_PROD_UNDO: UndoProcessor(view_boxes),
            TK_PROD_REFRESH_VIEW: RefreshViewProcessor(view_boxes),
            TK_PROD_GET_ORDER_MAX_PREP_TIME: max_prep_time_processor,
            TK_PROD_GET_ORDER_XML: GetOrderXmlProcessor(self.production_manager),
            TK_PROD_PURGE: PurgeOrdersProcessor(self.production_manager),
            TK_PROD_GET_PRINTERS_INFO: GetPrintersInfoProcessor(view_boxes),
            TK_PROD_GET_COURSES: GetCoursesProcessor(self.product_repository),
            TK_PROD_ORDER_PICKLIST: PickListOrdersProcessor(
                message_bus=self.message_bus,
                production_repository=self.production_repository,
            )
        }

        callbacks = [LoggingProcessorCallback(manager_logger)]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=event_processors,
            message_processors=message_processors,
            executor_factory=DefaultMessageProcessorExecutorFactory(callbacks),
        )

    def destroy_message_handler(
        self,
        message_handler,    # type: ProductionMessageHandlerBuilder
    ):
        # type: (...) -> None

        pass

    def build_message_handler(self, message, event):
        return self.message_handler

    def destroy_singletons(self):
        self.production_manager.stop()
        self.user_cache_thread.stop()
        self.production_repository.stop()
        self.clean_picklist_database_service.stop()
        self.product_connection.close()
        self.production_connection.close()

    def _build_product_repository(
        self,  # type: ProductionMessageHandlerBuilder
    ):
        # type: (...) -> ProductRepository

        self.product_connection = self._get_connection(
            connection_string=self.product_connection_string
        )

        return ProductRepository(
            connection=self.product_connection,
        )

    def _build_production_repository(
        self,   # type: ProductionMessageHandlerBuilder
    ):
        # type: (...) -> ProductionRepository

        self.production_connection = self._get_connection(
            connection_string=self.production_connection_string
        )
        synchronized_production_connection = SynchronizedDbApiTemplate(
            conn=self.production_connection,
        )
        return ProductionRepository(
            loggers=self.loggers,
            connection=synchronized_production_connection,
        )

    @staticmethod
    def _get_connection(
        connection_string,  # type:str
    ):
        # type: (...) -> sqlite3.Connection

        return sqlite3.connect(
            database=connection_string,
            check_same_thread=False,
        )



from __future__ import annotations

import email.utils
import mimetypes
import typing

_TYPE_FIELD_VALUE = typing.Union[str, bytes]
_TYPE_FIELD_VALUE_TUPLE = typing.Union[
    _TYPE_FIELD_VALUE,
    typing.Tuple[str, _TYPE_FIELD_VALUE],
    typing.Tuple[str, _TYPE_FIELD_VALUE, str],
]


def guess_content_type(
    filename: str | None, default: str = "application/octet-stream"
) -> str:
    """
    Guess the "Content-Type" of a file.

    :param filename:
        The filename to guess the "Content-Type" of using :mod:`mimetypes`.
    :param default:
        If no "Content-Type" can be guessed, default to `default`.
    """
    if filename:
        return mimetypes.guess_type(filename)[0] or default
    return default


def format_header_param_rfc2231(name: str, value: _TYPE_FIELD_VALUE) -> str:
    """
    Helper function to format and quote a single header parameter using the
    strategy defined in RFC 2231.

    Particularly useful for header parameters which might contain
    non-ASCII values, like file names. This follows
    `RFC 2388 Section 4.4 <https://tools.ietf.org/html/rfc2388#section-4.4>`_.

    :param name:
        The name of the parameter, a string expected to be ASCII only.
    :param value:
        The value of the parameter, provided as ``bytes`` or `str``.
    :returns:
        An RFC-2231-formatted unicode string.

    .. deprecated:: 2.0.0
        Will be removed in urllib3 v2.1.0. This is not valid for
        ``multipart/form-data`` header parameters.
    """
    import warnings

    warnings.warn(
        "'format_header_param_rfc2231' is deprecated and will be "
        "removed in urllib3 v2.1.0. This is not valid for "
        "multipart/form-data header parameters.",
        DeprecationWarning,
        stacklevel=2,
    )

    if isinstance(value, bytes):
        value = value.decode("utf-8")

    if not any(ch in value for ch in '"\\\r\n'):
        result = f'{name}="{value}"'
        try:
            result.encode("ascii")
        except (UnicodeEncodeError, UnicodeDecodeError):
            pass
        else:
            return result

    value = email.utils.encode_rfc2231(value, "utf-8")
    value = f"{name}*={value}"

    return value


def format_multipart_header_param(name: str, value: _TYPE_FIELD_VALUE) -> str:
    """
    Format and quote a single multipart header parameter.

    This follows the `WHATWG HTML Standard`_ as of 2021/06/10, matching
    the behavior of current browser and curl versions. Values are
    assumed to be UTF-8. The ``\\n``, ``\\r``, and ``"`` characters are
    percent encoded.

    .. _WHATWG HTML Standard:
        https://html.spec.whatwg.org/multipage/
        form-control-infrastructure.html#multipart-form-data

    :param name:
        The name of the parameter, an ASCII-only ``str``.
    :param value:
        The value of the parameter, a ``str`` or UTF-8 encoded
        ``bytes``.
    :returns:
        A string ``name="value"`` with the escaped value.

    .. versionchanged:: 2.0.0
        Matches the WHATWG HTML Standard as of 2021/06/10. Control
        characters are no longer percent encoded.

    .. versionchanged:: 2.0.0
        Renamed from ``format_header_param_html5`` and
        ``format_header_param``. The old names will be removed in
        urllib3 v2.1.0.
    """
    if isinstance(value, bytes):
        value = value.decode("utf-8")

    # percent encode \n \r "
    value = value.translate({10: "%0A", 13: "%0D", 34: "%22"})
    return f'{name}="{value}"'


def format_header_param_html5(name: str, value: _TYPE_FIELD_VALUE) -> str:
    """
    .. deprecated:: 2.0.0
        Renamed to :func:`format_multipart_header_param`. Will be
        removed in urllib3 v2.1.0.
    """
    import warnings

    warnings.warn(
        "'format_header_param_html5' has been renamed to "
        "'format_multipart_header_param'. The old name will be "
        "removed in urllib3 v2.1.0.",
        DeprecationWarning,
        stacklevel=2,
    )
    return format_multipart_header_param(name, value)


def format_header_param(name: str, value: _TYPE_FIELD_VALUE) -> str:
    """
    .. deprecated:: 2.0.0
        Renamed to :func:`format_multipart_header_param`. Will be
        removed in urllib3 v2.1.0.
    """
    import warnings

    warnings.warn(
        "'format_header_param' has been renamed to "
        "'format_multipart_header_param'. The old name will be "
        "removed in urllib3 v2.1.0.",
        DeprecationWarning,
        stacklevel=2,
    )
    return format_multipart_header_param(name, value)


class RequestField:
    """
    A data container for request body parameters.

    :param name:
        The name of this request field. Must be unicode.
    :param data:
        The data/value body.
    :param filename:
        An optional filename of the request field. Must be unicode.
    :param headers:
        An optional dict-like object of headers to initially use for the field.

    .. versionchanged:: 2.0.0
        The ``header_formatter`` parameter is deprecated and will
        be removed in urllib3 v2.1.0.
    """

    def __init__(
        self,
        name: str,
        data: _TYPE_FIELD_VALUE,
        filename: str | None = None,
        headers: typing.Mapping[str, str] | None = None,
        header_formatter: typing.Callable[[str, _TYPE_FIELD_VALUE], str] | None = None,
    ):
        self._name = name
        self._filename = filename
        self.data = data
        self.headers: dict[str, str | None] = {}
        if headers:
            self.headers = dict(headers)

        if header_formatter is not None:
            import warnings

            warnings.warn(
                "The 'header_formatter' parameter is deprecated and "
                "will be removed in urllib3 v2.1.0.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.header_formatter = header_formatter
        else:
            self.header_formatter = format_multipart_header_param

    @classmethod
    def from_tuples(
        cls,
        fieldname: str,
        value: _TYPE_FIELD_VALUE_TUPLE,
        header_formatter: typing.Callable[[str, _TYPE_FIELD_VALUE], str] | None = None,
    ) -> RequestField:
        """
        A :class:`~urllib3.fields.RequestField` factory from old-style tuple parameters.

        Supports constructing :class:`~urllib3.fields.RequestField` from
        parameter of key/value strings AND key/filetuple. A filetuple is a
        (filename, data, MIME type) tuple where the MIME type is optional.
        For example::

            'foo': 'bar',
            'fakefile': ('foofile.txt', 'contents of foofile'),
            'realfile': ('barfile.txt', open('realfile').read()),
            'typedfile': ('bazfile.bin', open('bazfile').read(), 'image/jpeg'),
            'nonamefile': 'contents of nonamefile field',

        Field names and filenames must be unicode.
        """
        filename: str | None
        content_type: str | None
        data: _TYPE_FIELD_VALUE

        if isinstance(value, tuple):
            if len(value) == 3:
                filename, data, content_type = typing.cast(
                    typing.Tuple[str, _TYPE_FIELD_VALUE, str], value
                )
            else:
                filename, data = typing.cast(
                    typing.Tuple[str, _TYPE_FIELD_VALUE], value
                )
                content_type = guess_content_type(filename)
        else:
            filename = None
            content_type = None
            data = value

        request_param = cls(
            fieldname, data, filename=filename, header_formatter=header_formatter
        )
        request_param.make_multipart(content_type=content_type)

        return request_param

    def _render_part(self, name: str, value: _TYPE_FIELD_VALUE) -> str:
        """
        Override this method to change how each multipart header
        parameter is formatted. By default, this calls
        :func:`format_multipart_header_param`.

        :param name:
            The name of the parameter, an ASCII-only ``str``.
        :param value:
            The value of the parameter, a ``str`` or UTF-8 encoded
            ``bytes``.

        :meta public:
        """
        return self.header_formatter(name, value)

    def _render_parts(
        self,
        header_parts: (
            dict[str, _TYPE_FIELD_VALUE | None]
            | typing.Sequence[tuple[str, _TYPE_FIELD_VALUE | None]]
        ),
    ) -> str:
        """
        Helper function to format and quote a single header.

        Useful for single headers that are composed of multiple items. E.g.,
        'Content-Disposition' fields.

        :param header_parts:
            A sequence of (k, v) tuples or a :class:`dict` of (k, v) to format
            as `k1="v1"; k2="v2"; ...`.
        """
        iterable: typing.Iterable[tuple[str, _TYPE_FIELD_VALUE | None]]

        parts = []
        if isinstance(header_parts, dict):
            iterable = header_parts.items()
        else:
            iterable = header_parts

        for name, value in iterable:
            if value is not None:
                parts.append(self._render_part(name, value))

        return "; ".join(parts)

    def render_headers(self) -> str:
        """
        Renders the headers for this request field.
        """
        lines = []

        sort_keys = ["Content-Disposition", "Content-Type", "Content-Location"]
        for sort_key in sort_keys:
            if self.headers.get(sort_key, False):
                lines.append(f"{sort_key}: {self.headers[sort_key]}")

        for header_name, header_value in self.headers.items():
            if header_name not in sort_keys:
                if header_value:
                    lines.append(f"{header_name}: {header_value}")

        lines.append("\r\n")
        return "\r\n".join(lines)

    def make_multipart(
        self,
        content_disposition: str | None = None,
        content_type: str | None = None,
        content_location: str | None = None,
    ) -> None:
        """
        Makes this request field into a multipart request field.

        This method overrides "Content-Disposition", "Content-Type" and
        "Content-Location" headers to the request parameter.

        :param content_disposition:
            The 'Content-Disposition' of the request body. Defaults to 'form-data'
        :param content_type:
            The 'Content-Type' of the request body.
        :param content_location:
            The 'Content-Location' of the request body.

        """
        content_disposition = (content_disposition or "form-data") + "; ".join(
            [
                "",
                self._render_parts(
                    (("name", self._name), ("filename", self._filename))
                ),
            ]
        )

        self.headers["Content-Disposition"] = content_disposition
        self.headers["Content-Type"] = content_type
        self.headers["Content-Location"] = content_location



from enum import IntEnum

from messagebus import TokenCreator, TokenPriority


MSGGRP_SLCTRL_ORDERMGR = "A"


def _create_order_token(number):
    return TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=MSGGRP_SLCTRL_ORDERMGR,
        number=number,
    )


# noinspection SpellCheckingInspection
class OrderApiTokens(IntEnum):

    TK_SLCTRL_OMGR_CHANGEMODIFIER = _create_order_token("13")
    TK_SLCTRL_OMGR_CHANGEQTY = _create_order_token("E")
    TK_SLCTRL_OMGR_CLEAROPTION = _create_order_token("7")
    TK_SLCTRL_OMGR_DOOPTION = _create_order_token("A")
    TK_SLCTRL_OMGR_DOSALE = _create_order_token("1")
    TK_SLCTRL_OMGR_DOTENDER = _create_order_token("10")
    TK_SLCTRL_OMGR_ORDERPICT = _create_order_token("2")
    TK_SLCTRL_OMGR_CREATEORDER = _create_order_token("18")
    TK_SLCTRL_OMGR_LISTORDERS = _create_order_token("D")
    TK_SLCTRL_OMGR_ORDERREOPEN = _create_order_token("F")
    TK_SLCTRL_OMGR_ORDERTOTAL = _create_order_token("6")
    TK_SLCTRL_OMGR_PULLORDERS = _create_order_token("35")
    TK_SLCTRL_OMGR_RECALLORDER = _create_order_token("C")
    TK_SLCTRL_OMGR_STOREORDER = _create_order_token("B")
    TK_SLCTRL_OMGR_VOIDLINE = _create_order_token("3")
    TK_SLCTRL_OMGR_VOIDORDER = _create_order_token("4")
    TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES = _create_order_token(number="28")
    TK_SLCTRL_OMGR_GET_CONFIG = _create_order_token("34")



# pesel.py - functions for handling Polish national identification numbers
# coding: utf-8
#
# Copyright (C) 2015 Dariusz Choruzy
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""PESEL (Polish national identification number).

The Powszechny Elektroniczny System Ewidencji Ludności (PESEL, Universal
Electronic System for Registration of the Population) is a 11-digit Polish
national identification number. The number consists of the date of birth,
a serial number, the person's gender and a check digit.


>>> validate('44051401359')
'44051401359'
>>> validate('44051401358')  # incorrect check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('02381307589')  # invalid birth date
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> get_birth_date('02122401358')
datetime.date(1902, 12, 24)
>>> get_gender('02122401358')
'M'
>>> get_birth_date('02211307589')
datetime.date(2002, 1, 13)
>>> get_gender('02211307589')
'F'
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def get_birth_date(number):
    """Split the date parts from the number and return the birth date."""
    number = compact(number)
    year = int(number[0:2])
    month = int(number[2:4])
    day = int(number[4:6])
    year += {
        0: 1900,
        1: 2000,
        2: 2100,
        3: 2200,
        4: 1800,
    }[month // 20]
    month = month % 20
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    if number[9] in '02468':  # even
        return 'F'
    else:  # odd: 13579
        return 'M'


def calc_check_digit(number):
    """Calculate the check digit for organisations. The number passed
    should not have the check digit included."""
    weights = (1, 3, 7, 9, 1, 3, 7, 9, 1, 3)
    check = sum(w * int(n) for w, n in zip(weights, number))
    return str((10 - check) % 10)


def validate(number):
    """Check if the number is a valid national identification number. This
    checks the length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    get_birth_date(number)
    return number


def is_valid(number):
    """Check if the number is a valid national identification number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from ._MeanTimeStatisticBox import MeanTimeStatisticBox
from ._OrdersCounterStatisticBox import OrdersCounterStatisticBox
from ._StatisticBox import StatisticBox
from ._TimeThresholdStatisticBox import TimeThresholdStatisticBox



#
# The Python Imaging Library.
# $Id$
#
# PDF (Acrobat) file handling
#
# History:
# 1996-07-16 fl   Created
# 1997-01-18 fl   Fixed header
# 2004-02-21 fl   Fixes for 1/L/CMYK images, etc.
# 2004-02-24 fl   Fixes for 1 and P images.
#
# Copyright (c) 1997-2004 by Secret Labs AB.  All rights reserved.
# Copyright (c) 1996-1997 by Fredrik Lundh.
#
# See the README file for information on usage and redistribution.
#

##
# Image plugin for PDF images (output only).
##
from __future__ import annotations

import io
import math
import os
import time

from . import Image, ImageFile, ImageSequence, PdfParser, __version__, features

#
# --------------------------------------------------------------------

# object ids:
#  1. catalogue
#  2. pages
#  3. image
#  4. page
#  5. page contents


def _save_all(im, fp, filename):
    _save(im, fp, filename, save_all=True)


##
# (Internal) Image save plugin for the PDF format.


def _write_image(im, filename, existing_pdf, image_refs):
    # FIXME: Should replace ASCIIHexDecode with RunLengthDecode
    # (packbits) or LZWDecode (tiff/lzw compression).  Note that
    # PDF 1.2 also supports Flatedecode (zip compression).

    params = None
    decode = None

    #
    # Get image characteristics

    width, height = im.size

    dict_obj = {"BitsPerComponent": 8}
    if im.mode == "1":
        if features.check("libtiff"):
            filter = "CCITTFaxDecode"
            dict_obj["BitsPerComponent"] = 1
            params = PdfParser.PdfArray(
                [
                    PdfParser.PdfDict(
                        {
                            "K": -1,
                            "BlackIs1": True,
                            "Columns": width,
                            "Rows": height,
                        }
                    )
                ]
            )
        else:
            filter = "DCTDecode"
        dict_obj["ColorSpace"] = PdfParser.PdfName("DeviceGray")
        procset = "ImageB"  # grayscale
    elif im.mode == "L":
        filter = "DCTDecode"
        # params = f"<< /Predictor 15 /Columns {width-2} >>"
        dict_obj["ColorSpace"] = PdfParser.PdfName("DeviceGray")
        procset = "ImageB"  # grayscale
    elif im.mode == "LA":
        filter = "JPXDecode"
        # params = f"<< /Predictor 15 /Columns {width-2} >>"
        procset = "ImageB"  # grayscale
        dict_obj["SMaskInData"] = 1
    elif im.mode == "P":
        filter = "ASCIIHexDecode"
        palette = im.getpalette()
        dict_obj["ColorSpace"] = [
            PdfParser.PdfName("Indexed"),
            PdfParser.PdfName("DeviceRGB"),
            len(palette) // 3 - 1,
            PdfParser.PdfBinary(palette),
        ]
        procset = "ImageI"  # indexed color

        if "transparency" in im.info:
            smask = im.convert("LA").getchannel("A")
            smask.encoderinfo = {}

            image_ref = _write_image(smask, filename, existing_pdf, image_refs)[0]
            dict_obj["SMask"] = image_ref
    elif im.mode == "RGB":
        filter = "DCTDecode"
        dict_obj["ColorSpace"] = PdfParser.PdfName("DeviceRGB")
        procset = "ImageC"  # color images
    elif im.mode == "RGBA":
        filter = "JPXDecode"
        procset = "ImageC"  # color images
        dict_obj["SMaskInData"] = 1
    elif im.mode == "CMYK":
        filter = "DCTDecode"
        dict_obj["ColorSpace"] = PdfParser.PdfName("DeviceCMYK")
        procset = "ImageC"  # color images
        decode = [1, 0, 1, 0, 1, 0, 1, 0]
    else:
        msg = f"cannot save mode {im.mode}"
        raise ValueError(msg)

    #
    # image

    op = io.BytesIO()

    if filter == "ASCIIHexDecode":
        ImageFile._save(im, op, [("hex", (0, 0) + im.size, 0, im.mode)])
    elif filter == "CCITTFaxDecode":
        im.save(
            op,
            "TIFF",
            compression="group4",
            # use a single strip
            strip_size=math.ceil(width / 8) * height,
        )
    elif filter == "DCTDecode":
        Image.SAVE["JPEG"](im, op, filename)
    elif filter == "JPXDecode":
        del dict_obj["BitsPerComponent"]
        Image.SAVE["JPEG2000"](im, op, filename)
    else:
        msg = f"unsupported PDF filter ({filter})"
        raise ValueError(msg)

    stream = op.getvalue()
    if filter == "CCITTFaxDecode":
        stream = stream[8:]
        filter = PdfParser.PdfArray([PdfParser.PdfName(filter)])
    else:
        filter = PdfParser.PdfName(filter)

    image_ref = image_refs.pop(0)
    existing_pdf.write_obj(
        image_ref,
        stream=stream,
        Type=PdfParser.PdfName("XObject"),
        Subtype=PdfParser.PdfName("Image"),
        Width=width,  # * 72.0 / x_resolution,
        Height=height,  # * 72.0 / y_resolution,
        Filter=filter,
        Decode=decode,
        DecodeParms=params,
        **dict_obj,
    )

    return image_ref, procset


def _save(im, fp, filename, save_all=False):
    is_appending = im.encoderinfo.get("append", False)
    if is_appending:
        existing_pdf = PdfParser.PdfParser(f=fp, filename=filename, mode="r+b")
    else:
        existing_pdf = PdfParser.PdfParser(f=fp, filename=filename, mode="w+b")

    dpi = im.encoderinfo.get("dpi")
    if dpi:
        x_resolution = dpi[0]
        y_resolution = dpi[1]
    else:
        x_resolution = y_resolution = im.encoderinfo.get("resolution", 72.0)

    info = {
        "title": None
        if is_appending
        else os.path.splitext(os.path.basename(filename))[0],
        "author": None,
        "subject": None,
        "keywords": None,
        "creator": None,
        "producer": None,
        "creationDate": None if is_appending else time.gmtime(),
        "modDate": None if is_appending else time.gmtime(),
    }
    for k, default in info.items():
        v = im.encoderinfo.get(k) if k in im.encoderinfo else default
        if v:
            existing_pdf.info[k[0].upper() + k[1:]] = v

    #
    # make sure image data is available
    im.load()

    existing_pdf.start_writing()
    existing_pdf.write_header()
    existing_pdf.write_comment(f"created by Pillow {__version__} PDF driver")

    #
    # pages
    ims = [im]
    if save_all:
        append_images = im.encoderinfo.get("append_images", [])
        for append_im in append_images:
            append_im.encoderinfo = im.encoderinfo.copy()
            ims.append(append_im)
    number_of_pages = 0
    image_refs = []
    page_refs = []
    contents_refs = []
    for im in ims:
        im_number_of_pages = 1
        if save_all:
            try:
                im_number_of_pages = im.n_frames
            except AttributeError:
                # Image format does not have n_frames.
                # It is a single frame image
                pass
        number_of_pages += im_number_of_pages
        for i in range(im_number_of_pages):
            image_refs.append(existing_pdf.next_object_id(0))
            if im.mode == "P" and "transparency" in im.info:
                image_refs.append(existing_pdf.next_object_id(0))

            page_refs.append(existing_pdf.next_object_id(0))
            contents_refs.append(existing_pdf.next_object_id(0))
            existing_pdf.pages.append(page_refs[-1])

    #
    # catalog and list of pages
    existing_pdf.write_catalog()

    page_number = 0
    for im_sequence in ims:
        im_pages = ImageSequence.Iterator(im_sequence) if save_all else [im_sequence]
        for im in im_pages:
            image_ref, procset = _write_image(im, filename, existing_pdf, image_refs)

            #
            # page

            existing_pdf.write_page(
                page_refs[page_number],
                Resources=PdfParser.PdfDict(
                    ProcSet=[PdfParser.PdfName("PDF"), PdfParser.PdfName(procset)],
                    XObject=PdfParser.PdfDict(image=image_ref),
                ),
                MediaBox=[
                    0,
                    0,
                    im.width * 72.0 / x_resolution,
                    im.height * 72.0 / y_resolution,
                ],
                Contents=contents_refs[page_number],
            )

            #
            # page contents

            page_contents = b"q %f 0 0 %f 0 0 cm /image Do Q\n" % (
                im.width * 72.0 / x_resolution,
                im.height * 72.0 / y_resolution,
            )

            existing_pdf.write_obj(contents_refs[page_number], stream=page_contents)

            page_number += 1

    #
    # trailer
    existing_pdf.write_xref_and_trailer()
    if hasattr(fp, "flush"):
        fp.flush()
    existing_pdf.close()


#
# --------------------------------------------------------------------


Image.register_save("PDF", _save)
Image.register_save_all("PDF", _save_all)

Image.register_extension("PDF", ".pdf")

Image.register_mime("PDF", "application/pdf")



#------------------------------------------------------------------------------
# pycparser: c_generator.py
#
# C code generator from pycparser AST nodes.
#
# Eli Bendersky [https://eli.thegreenplace.net/]
# License: BSD
#------------------------------------------------------------------------------
from . import c_ast


class CGenerator(object):
    """ Uses the same visitor pattern as c_ast.NodeVisitor, but modified to
        return a value from each visit method, using string accumulation in
        generic_visit.
    """
    def __init__(self, reduce_parentheses=False):
        """ Constructs C-code generator

            reduce_parentheses:
                if True, eliminates needless parentheses on binary operators
        """
        # Statements start with indentation of self.indent_level spaces, using
        # the _make_indent method.
        self.indent_level = 0
        self.reduce_parentheses = reduce_parentheses

    def _make_indent(self):
        return ' ' * self.indent_level

    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        return getattr(self, method, self.generic_visit)(node)

    def generic_visit(self, node):
        if node is None:
            return ''
        else:
            return ''.join(self.visit(c) for c_name, c in node.children())

    def visit_Constant(self, n):
        return n.value

    def visit_ID(self, n):
        return n.name

    def visit_Pragma(self, n):
        ret = '#pragma'
        if n.string:
            ret += ' ' + n.string
        return ret

    def visit_ArrayRef(self, n):
        arrref = self._parenthesize_unless_simple(n.name)
        return arrref + '[' + self.visit(n.subscript) + ']'

    def visit_StructRef(self, n):
        sref = self._parenthesize_unless_simple(n.name)
        return sref + n.type + self.visit(n.field)

    def visit_FuncCall(self, n):
        fref = self._parenthesize_unless_simple(n.name)
        return fref + '(' + self.visit(n.args) + ')'

    def visit_UnaryOp(self, n):
        if n.op == 'sizeof':
            # Always parenthesize the argument of sizeof since it can be
            # a name.
            return 'sizeof(%s)' % self.visit(n.expr)
        else:
            operand = self._parenthesize_unless_simple(n.expr)
            if n.op == 'p++':
                return '%s++' % operand
            elif n.op == 'p--':
                return '%s--' % operand
            else:
                return '%s%s' % (n.op, operand)

    # Precedence map of binary operators:
    precedence_map = {
        # Should be in sync with c_parser.CParser.precedence
        # Higher numbers are stronger binding
        '||': 0,  # weakest binding
        '&&': 1,
        '|': 2,
        '^': 3,
        '&': 4,
        '==': 5, '!=': 5,
        '>': 6, '>=': 6, '<': 6, '<=': 6,
        '>>': 7, '<<': 7,
        '+': 8, '-': 8,
        '*': 9, '/': 9, '%': 9  # strongest binding
    }

    def visit_BinaryOp(self, n):
        # Note: all binary operators are left-to-right associative
        #
        # If `n.left.op` has a stronger or equally binding precedence in
        # comparison to `n.op`, no parenthesis are needed for the left:
        # e.g., `(a*b) + c` is equivalent to `a*b + c`, as well as
        #       `(a+b) - c` is equivalent to `a+b - c` (same precedence).
        # If the left operator is weaker binding than the current, then
        # parentheses are necessary:
        # e.g., `(a+b) * c` is NOT equivalent to `a+b * c`.
        lval_str = self._parenthesize_if(
            n.left,
            lambda d: not (self._is_simple_node(d) or
                      self.reduce_parentheses and isinstance(d, c_ast.BinaryOp) and
                      self.precedence_map[d.op] >= self.precedence_map[n.op]))
        # If `n.right.op` has a stronger -but not equal- binding precedence,
        # parenthesis can be omitted on the right:
        # e.g., `a + (b*c)` is equivalent to `a + b*c`.
        # If the right operator is weaker or equally binding, then parentheses
        # are necessary:
        # e.g., `a * (b+c)` is NOT equivalent to `a * b+c` and
        #       `a - (b+c)` is NOT equivalent to `a - b+c` (same precedence).
        rval_str = self._parenthesize_if(
            n.right,
            lambda d: not (self._is_simple_node(d) or
                      self.reduce_parentheses and isinstance(d, c_ast.BinaryOp) and
                      self.precedence_map[d.op] > self.precedence_map[n.op]))
        return '%s %s %s' % (lval_str, n.op, rval_str)

    def visit_Assignment(self, n):
        rval_str = self._parenthesize_if(
                            n.rvalue,
                            lambda n: isinstance(n, c_ast.Assignment))
        return '%s %s %s' % (self.visit(n.lvalue), n.op, rval_str)

    def visit_IdentifierType(self, n):
        return ' '.join(n.names)

    def _visit_expr(self, n):
        if isinstance(n, c_ast.InitList):
            return '{' + self.visit(n) + '}'
        elif isinstance(n, c_ast.ExprList):
            return '(' + self.visit(n) + ')'
        else:
            return self.visit(n)

    def visit_Decl(self, n, no_type=False):
        # no_type is used when a Decl is part of a DeclList, where the type is
        # explicitly only for the first declaration in a list.
        #
        s = n.name if no_type else self._generate_decl(n)
        if n.bitsize: s += ' : ' + self.visit(n.bitsize)
        if n.init:
            s += ' = ' + self._visit_expr(n.init)
        return s

    def visit_DeclList(self, n):
        s = self.visit(n.decls[0])
        if len(n.decls) > 1:
            s += ', ' + ', '.join(self.visit_Decl(decl, no_type=True)
                                    for decl in n.decls[1:])
        return s

    def visit_Typedef(self, n):
        s = ''
        if n.storage: s += ' '.join(n.storage) + ' '
        s += self._generate_type(n.type)
        return s

    def visit_Cast(self, n):
        s = '(' + self._generate_type(n.to_type, emit_declname=False) + ')'
        return s + ' ' + self._parenthesize_unless_simple(n.expr)

    def visit_ExprList(self, n):
        visited_subexprs = []
        for expr in n.exprs:
            visited_subexprs.append(self._visit_expr(expr))
        return ', '.join(visited_subexprs)

    def visit_InitList(self, n):
        visited_subexprs = []
        for expr in n.exprs:
            visited_subexprs.append(self._visit_expr(expr))
        return ', '.join(visited_subexprs)

    def visit_Enum(self, n):
        return self._generate_struct_union_enum(n, name='enum')

    def visit_Alignas(self, n):
        return '_Alignas({})'.format(self.visit(n.alignment))

    def visit_Enumerator(self, n):
        if not n.value:
            return '{indent}{name},\n'.format(
                indent=self._make_indent(),
                name=n.name,
            )
        else:
            return '{indent}{name} = {value},\n'.format(
                indent=self._make_indent(),
                name=n.name,
                value=self.visit(n.value),
            )

    def visit_FuncDef(self, n):
        decl = self.visit(n.decl)
        self.indent_level = 0
        body = self.visit(n.body)
        if n.param_decls:
            knrdecls = ';\n'.join(self.visit(p) for p in n.param_decls)
            return decl + '\n' + knrdecls + ';\n' + body + '\n'
        else:
            return decl + '\n' + body + '\n'

    def visit_FileAST(self, n):
        s = ''
        for ext in n.ext:
            if isinstance(ext, c_ast.FuncDef):
                s += self.visit(ext)
            elif isinstance(ext, c_ast.Pragma):
                s += self.visit(ext) + '\n'
            else:
                s += self.visit(ext) + ';\n'
        return s

    def visit_Compound(self, n):
        s = self._make_indent() + '{\n'
        self.indent_level += 2
        if n.block_items:
            s += ''.join(self._generate_stmt(stmt) for stmt in n.block_items)
        self.indent_level -= 2
        s += self._make_indent() + '}\n'
        return s

    def visit_CompoundLiteral(self, n):
        return '(' + self.visit(n.type) + '){' + self.visit(n.init) + '}'


    def visit_EmptyStatement(self, n):
        return ';'

    def visit_ParamList(self, n):
        return ', '.join(self.visit(param) for param in n.params)

    def visit_Return(self, n):
        s = 'return'
        if n.expr: s += ' ' + self.visit(n.expr)
        return s + ';'

    def visit_Break(self, n):
        return 'break;'

    def visit_Continue(self, n):
        return 'continue;'

    def visit_TernaryOp(self, n):
        s  = '(' + self._visit_expr(n.cond) + ') ? '
        s += '(' + self._visit_expr(n.iftrue) + ') : '
        s += '(' + self._visit_expr(n.iffalse) + ')'
        return s

    def visit_If(self, n):
        s = 'if ('
        if n.cond: s += self.visit(n.cond)
        s += ')\n'
        s += self._generate_stmt(n.iftrue, add_indent=True)
        if n.iffalse:
            s += self._make_indent() + 'else\n'
            s += self._generate_stmt(n.iffalse, add_indent=True)
        return s

    def visit_For(self, n):
        s = 'for ('
        if n.init: s += self.visit(n.init)
        s += ';'
        if n.cond: s += ' ' + self.visit(n.cond)
        s += ';'
        if n.next: s += ' ' + self.visit(n.next)
        s += ')\n'
        s += self._generate_stmt(n.stmt, add_indent=True)
        return s

    def visit_While(self, n):
        s = 'while ('
        if n.cond: s += self.visit(n.cond)
        s += ')\n'
        s += self._generate_stmt(n.stmt, add_indent=True)
        return s

    def visit_DoWhile(self, n):
        s = 'do\n'
        s += self._generate_stmt(n.stmt, add_indent=True)
        s += self._make_indent() + 'while ('
        if n.cond: s += self.visit(n.cond)
        s += ');'
        return s

    def visit_StaticAssert(self, n):
        s = '_Static_assert('
        s += self.visit(n.cond)
        if n.message:
            s += ','
            s += self.visit(n.message)
        s += ')'
        return s

    def visit_Switch(self, n):
        s = 'switch (' + self.visit(n.cond) + ')\n'
        s += self._generate_stmt(n.stmt, add_indent=True)
        return s

    def visit_Case(self, n):
        s = 'case ' + self.visit(n.expr) + ':\n'
        for stmt in n.stmts:
            s += self._generate_stmt(stmt, add_indent=True)
        return s

    def visit_Default(self, n):
        s = 'default:\n'
        for stmt in n.stmts:
            s += self._generate_stmt(stmt, add_indent=True)
        return s

    def visit_Label(self, n):
        return n.name + ':\n' + self._generate_stmt(n.stmt)

    def visit_Goto(self, n):
        return 'goto ' + n.name + ';'

    def visit_EllipsisParam(self, n):
        return '...'

    def visit_Struct(self, n):
        return self._generate_struct_union_enum(n, 'struct')

    def visit_Typename(self, n):
        return self._generate_type(n.type)

    def visit_Union(self, n):
        return self._generate_struct_union_enum(n, 'union')

    def visit_NamedInitializer(self, n):
        s = ''
        for name in n.name:
            if isinstance(name, c_ast.ID):
                s += '.' + name.name
            else:
                s += '[' + self.visit(name) + ']'
        s += ' = ' + self._visit_expr(n.expr)
        return s

    def visit_FuncDecl(self, n):
        return self._generate_type(n)

    def visit_ArrayDecl(self, n):
        return self._generate_type(n, emit_declname=False)

    def visit_TypeDecl(self, n):
        return self._generate_type(n, emit_declname=False)

    def visit_PtrDecl(self, n):
        return self._generate_type(n, emit_declname=False)

    def _generate_struct_union_enum(self, n, name):
        """ Generates code for structs, unions, and enums. name should be
            'struct', 'union', or 'enum'.
        """
        if name in ('struct', 'union'):
            members = n.decls
            body_function = self._generate_struct_union_body
        else:
            assert name == 'enum'
            members = None if n.values is None else n.values.enumerators
            body_function = self._generate_enum_body
        s = name + ' ' + (n.name or '')
        if members is not None:
            # None means no members
            # Empty sequence means an empty list of members
            s += '\n'
            s += self._make_indent()
            self.indent_level += 2
            s += '{\n'
            s += body_function(members)
            self.indent_level -= 2
            s += self._make_indent() + '}'
        return s

    def _generate_struct_union_body(self, members):
        return ''.join(self._generate_stmt(decl) for decl in members)

    def _generate_enum_body(self, members):
        # `[:-2] + '\n'` removes the final `,` from the enumerator list
        return ''.join(self.visit(value) for value in members)[:-2] + '\n'

    def _generate_stmt(self, n, add_indent=False):
        """ Generation from a statement node. This method exists as a wrapper
            for individual visit_* methods to handle different treatment of
            some statements in this context.
        """
        typ = type(n)
        if add_indent: self.indent_level += 2
        indent = self._make_indent()
        if add_indent: self.indent_level -= 2

        if typ in (
                c_ast.Decl, c_ast.Assignment, c_ast.Cast, c_ast.UnaryOp,
                c_ast.BinaryOp, c_ast.TernaryOp, c_ast.FuncCall, c_ast.ArrayRef,
                c_ast.StructRef, c_ast.Constant, c_ast.ID, c_ast.Typedef,
                c_ast.ExprList):
            # These can also appear in an expression context so no semicolon
            # is added to them automatically
            #
            return indent + self.visit(n) + ';\n'
        elif typ in (c_ast.Compound,):
            # No extra indentation required before the opening brace of a
            # compound - because it consists of multiple lines it has to
            # compute its own indentation.
            #
            return self.visit(n)
        elif typ in (c_ast.If,):
            return indent + self.visit(n)
        else:
            return indent + self.visit(n) + '\n'

    def _generate_decl(self, n):
        """ Generation from a Decl node.
        """
        s = ''
        if n.funcspec: s = ' '.join(n.funcspec) + ' '
        if n.storage: s += ' '.join(n.storage) + ' '
        if n.align: s += self.visit(n.align[0]) + ' '
        s += self._generate_type(n.type)
        return s

    def _generate_type(self, n, modifiers=[], emit_declname = True):
        """ Recursive generation from a type node. n is the type node.
            modifiers collects the PtrDecl, ArrayDecl and FuncDecl modifiers
            encountered on the way down to a TypeDecl, to allow proper
            generation from it.
        """
        typ = type(n)
        #~ print(n, modifiers)

        if typ == c_ast.TypeDecl:
            s = ''
            if n.quals: s += ' '.join(n.quals) + ' '
            s += self.visit(n.type)

            nstr = n.declname if n.declname and emit_declname else ''
            # Resolve modifiers.
            # Wrap in parens to distinguish pointer to array and pointer to
            # function syntax.
            #
            for i, modifier in enumerate(modifiers):
                if isinstance(modifier, c_ast.ArrayDecl):
                    if (i != 0 and
                        isinstance(modifiers[i - 1], c_ast.PtrDecl)):
                            nstr = '(' + nstr + ')'
                    nstr += '['
                    if modifier.dim_quals:
                        nstr += ' '.join(modifier.dim_quals) + ' '
                    nstr += self.visit(modifier.dim) + ']'
                elif isinstance(modifier, c_ast.FuncDecl):
                    if (i != 0 and
                        isinstance(modifiers[i - 1], c_ast.PtrDecl)):
                            nstr = '(' + nstr + ')'
                    nstr += '(' + self.visit(modifier.args) + ')'
                elif isinstance(modifier, c_ast.PtrDecl):
                    if modifier.quals:
                        nstr = '* %s%s' % (' '.join(modifier.quals),
                                           ' ' + nstr if nstr else '')
                    else:
                        nstr = '*' + nstr
            if nstr: s += ' ' + nstr
            return s
        elif typ == c_ast.Decl:
            return self._generate_decl(n.type)
        elif typ == c_ast.Typename:
            return self._generate_type(n.type, emit_declname = emit_declname)
        elif typ == c_ast.IdentifierType:
            return ' '.join(n.names) + ' '
        elif typ in (c_ast.ArrayDecl, c_ast.PtrDecl, c_ast.FuncDecl):
            return self._generate_type(n.type, modifiers + [n],
                                       emit_declname = emit_declname)
        else:
            return self.visit(n)

    def _parenthesize_if(self, n, condition):
        """ Visits 'n' and returns its string representation, parenthesized
            if the condition function applied to the node returns True.
        """
        s = self._visit_expr(n)
        if condition(n):
            return '(' + s + ')'
        else:
            return s

    def _parenthesize_unless_simple(self, n):
        """ Common use case for _parenthesize_if
        """
        return self._parenthesize_if(n, lambda d: not self._is_simple_node(d))

    def _is_simple_node(self, n):
        """ Returns True for nodes that are "simple" - i.e. nodes that always
            have higher precedence than operators.
        """
        return isinstance(n, (c_ast.Constant, c_ast.ID, c_ast.ArrayRef,
                              c_ast.StructRef, c_ast.FuncCall))



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

from cryptography.hazmat._oid import (
    AttributeOID,
    AuthorityInformationAccessOID,
    CertificatePoliciesOID,
    CRLEntryExtensionOID,
    ExtendedKeyUsageOID,
    ExtensionOID,
    NameOID,
    ObjectIdentifier,
    OCSPExtensionOID,
    PublicKeyAlgorithmOID,
    SignatureAlgorithmOID,
    SubjectInformationAccessOID,
)

__all__ = [
    "AttributeOID",
    "AuthorityInformationAccessOID",
    "CRLEntryExtensionOID",
    "CertificatePoliciesOID",
    "ExtendedKeyUsageOID",
    "ExtensionOID",
    "NameOID",
    "OCSPExtensionOID",
    "ObjectIdentifier",
    "PublicKeyAlgorithmOID",
    "SignatureAlgorithmOID",
    "SubjectInformationAccessOID",
]



# mypy: ignore-errors

from sqlalchemy import bindparam
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import testing
from sqlalchemy import text
from sqlalchemy.testing import eq_
from sqlalchemy.testing import fixtures


class RowCountTest(fixtures.TablesTest):
    """test rowcount functionality"""

    __requires__ = ("sane_rowcount",)
    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table(
            "employees",
            metadata,
            Column(
                "employee_id",
                Integer,
                autoincrement=False,
                primary_key=True,
            ),
            Column("name", String(50)),
            Column("department", String(1)),
        )

    @classmethod
    def insert_data(cls, connection):
        cls.data = data = [
            ("Angela", "A"),
            ("Andrew", "A"),
            ("Anand", "A"),
            ("Bob", "B"),
            ("Bobette", "B"),
            ("Buffy", "B"),
            ("Charlie", "C"),
            ("Cynthia", "C"),
            ("Chris", "C"),
        ]

        employees_table = cls.tables.employees
        connection.execute(
            employees_table.insert(),
            [
                {"employee_id": i, "name": n, "department": d}
                for i, (n, d) in enumerate(data)
            ],
        )

    def test_basic(self, connection):
        employees_table = self.tables.employees
        s = select(
            employees_table.c.name, employees_table.c.department
        ).order_by(employees_table.c.employee_id)
        rows = connection.execute(s).fetchall()

        eq_(rows, self.data)

    @testing.variation("statement", ["update", "delete", "insert", "select"])
    @testing.variation("close_first", [True, False])
    def test_non_rowcount_scenarios_no_raise(
        self, connection, statement, close_first
    ):
        employees_table = self.tables.employees

        # WHERE matches 3, 3 rows changed
        department = employees_table.c.department

        if statement.update:
            r = connection.execute(
                employees_table.update().where(department == "C"),
                {"department": "Z"},
            )
        elif statement.delete:
            r = connection.execute(
                employees_table.delete().where(department == "C"),
                {"department": "Z"},
            )
        elif statement.insert:
            r = connection.execute(
                employees_table.insert(),
                [
                    {"employee_id": 25, "name": "none 1", "department": "X"},
                    {"employee_id": 26, "name": "none 2", "department": "Z"},
                    {"employee_id": 27, "name": "none 3", "department": "Z"},
                ],
            )
        elif statement.select:
            s = select(
                employees_table.c.name, employees_table.c.department
            ).where(employees_table.c.department == "C")
            r = connection.execute(s)
            r.all()
        else:
            statement.fail()

        if close_first:
            r.close()

        assert r.rowcount in (-1, 3)

    def test_update_rowcount1(self, connection):
        employees_table = self.tables.employees

        # WHERE matches 3, 3 rows changed
        department = employees_table.c.department
        r = connection.execute(
            employees_table.update().where(department == "C"),
            {"department": "Z"},
        )
        assert r.rowcount == 3

    def test_update_rowcount2(self, connection):
        employees_table = self.tables.employees

        # WHERE matches 3, 0 rows changed
        department = employees_table.c.department

        r = connection.execute(
            employees_table.update().where(department == "C"),
            {"department": "C"},
        )
        eq_(r.rowcount, 3)

    @testing.variation("implicit_returning", [True, False])
    @testing.variation(
        "dml",
        [
            ("update", testing.requires.update_returning),
            ("delete", testing.requires.delete_returning),
        ],
    )
    def test_update_delete_rowcount_return_defaults(
        self, connection, implicit_returning, dml
    ):
        """note this test should succeed for all RETURNING backends
        as of 2.0.  In
        Idf28379f8705e403a3c6a937f6a798a042ef2540 we changed rowcount to use
        len(rows) when we have implicit returning

        """

        if implicit_returning:
            employees_table = self.tables.employees
        else:
            employees_table = Table(
                "employees",
                MetaData(),
                Column(
                    "employee_id",
                    Integer,
                    autoincrement=False,
                    primary_key=True,
                ),
                Column("name", String(50)),
                Column("department", String(1)),
                implicit_returning=False,
            )

        department = employees_table.c.department

        if dml.update:
            stmt = (
                employees_table.update()
                .where(department == "C")
                .values(name=employees_table.c.department + "Z")
                .return_defaults()
            )
        elif dml.delete:
            stmt = (
                employees_table.delete()
                .where(department == "C")
                .return_defaults()
            )
        else:
            dml.fail()

        r = connection.execute(stmt)
        eq_(r.rowcount, 3)

    def test_raw_sql_rowcount(self, connection):
        # test issue #3622, make sure eager rowcount is called for text
        result = connection.exec_driver_sql(
            "update employees set department='Z' where department='C'"
        )
        eq_(result.rowcount, 3)

    def test_text_rowcount(self, connection):
        # test issue #3622, make sure eager rowcount is called for text
        result = connection.execute(
            text("update employees set department='Z' " "where department='C'")
        )
        eq_(result.rowcount, 3)

    def test_delete_rowcount(self, connection):
        employees_table = self.tables.employees

        # WHERE matches 3, 3 rows deleted
        department = employees_table.c.department
        r = connection.execute(
            employees_table.delete().where(department == "C")
        )
        eq_(r.rowcount, 3)

    @testing.requires.sane_multi_rowcount
    def test_multi_update_rowcount(self, connection):
        employees_table = self.tables.employees
        stmt = (
            employees_table.update()
            .where(employees_table.c.name == bindparam("emp_name"))
            .values(department="C")
        )

        r = connection.execute(
            stmt,
            [
                {"emp_name": "Bob"},
                {"emp_name": "Cynthia"},
                {"emp_name": "nonexistent"},
            ],
        )

        eq_(r.rowcount, 2)

    @testing.requires.sane_multi_rowcount
    def test_multi_delete_rowcount(self, connection):
        employees_table = self.tables.employees

        stmt = employees_table.delete().where(
            employees_table.c.name == bindparam("emp_name")
        )

        r = connection.execute(
            stmt,
            [
                {"emp_name": "Bob"},
                {"emp_name": "Cynthia"},
                {"emp_name": "nonexistent"},
            ],
        )

        eq_(r.rowcount, 2)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import abc
import typing
from math import gcd

from cryptography.hazmat.bindings._rust import openssl as rust_openssl
from cryptography.hazmat.primitives import _serialization, hashes
from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric import utils as asym_utils


class RSAPrivateKey(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def decrypt(self, ciphertext: bytes, padding: AsymmetricPadding) -> bytes:
        """
        Decrypts the provided ciphertext.
        """

    @property
    @abc.abstractmethod
    def key_size(self) -> int:
        """
        The bit length of the public modulus.
        """

    @abc.abstractmethod
    def public_key(self) -> RSAPublicKey:
        """
        The RSAPublicKey associated with this private key.
        """

    @abc.abstractmethod
    def sign(
        self,
        data: bytes,
        padding: AsymmetricPadding,
        algorithm: asym_utils.Prehashed | hashes.HashAlgorithm,
    ) -> bytes:
        """
        Signs the data.
        """

    @abc.abstractmethod
    def private_numbers(self) -> RSAPrivateNumbers:
        """
        Returns an RSAPrivateNumbers.
        """

    @abc.abstractmethod
    def private_bytes(
        self,
        encoding: _serialization.Encoding,
        format: _serialization.PrivateFormat,
        encryption_algorithm: _serialization.KeySerializationEncryption,
    ) -> bytes:
        """
        Returns the key serialized as bytes.
        """


RSAPrivateKeyWithSerialization = RSAPrivateKey
RSAPrivateKey.register(rust_openssl.rsa.RSAPrivateKey)


class RSAPublicKey(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encrypt(self, plaintext: bytes, padding: AsymmetricPadding) -> bytes:
        """
        Encrypts the given plaintext.
        """

    @property
    @abc.abstractmethod
    def key_size(self) -> int:
        """
        The bit length of the public modulus.
        """

    @abc.abstractmethod
    def public_numbers(self) -> RSAPublicNumbers:
        """
        Returns an RSAPublicNumbers
        """

    @abc.abstractmethod
    def public_bytes(
        self,
        encoding: _serialization.Encoding,
        format: _serialization.PublicFormat,
    ) -> bytes:
        """
        Returns the key serialized as bytes.
        """

    @abc.abstractmethod
    def verify(
        self,
        signature: bytes,
        data: bytes,
        padding: AsymmetricPadding,
        algorithm: asym_utils.Prehashed | hashes.HashAlgorithm,
    ) -> None:
        """
        Verifies the signature of the data.
        """

    @abc.abstractmethod
    def recover_data_from_signature(
        self,
        signature: bytes,
        padding: AsymmetricPadding,
        algorithm: hashes.HashAlgorithm | None,
    ) -> bytes:
        """
        Recovers the original data from the signature.
        """

    @abc.abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Checks equality.
        """


RSAPublicKeyWithSerialization = RSAPublicKey
RSAPublicKey.register(rust_openssl.rsa.RSAPublicKey)

RSAPrivateNumbers = rust_openssl.rsa.RSAPrivateNumbers
RSAPublicNumbers = rust_openssl.rsa.RSAPublicNumbers


def generate_private_key(
    public_exponent: int,
    key_size: int,
    backend: typing.Any = None,
) -> RSAPrivateKey:
    _verify_rsa_parameters(public_exponent, key_size)
    return rust_openssl.rsa.generate_private_key(public_exponent, key_size)


def _verify_rsa_parameters(public_exponent: int, key_size: int) -> None:
    if public_exponent not in (3, 65537):
        raise ValueError(
            "public_exponent must be either 3 (for legacy compatibility) or "
            "65537. Almost everyone should choose 65537 here!"
        )

    if key_size < 1024:
        raise ValueError("key_size must be at least 1024-bits.")


def _modinv(e: int, m: int) -> int:
    """
    Modular Multiplicative Inverse. Returns x such that: (x*e) mod m == 1
    """
    x1, x2 = 1, 0
    a, b = e, m
    while b > 0:
        q, r = divmod(a, b)
        xn = x1 - q * x2
        a, b, x1, x2 = b, r, x2, xn
    return x1 % m


def rsa_crt_iqmp(p: int, q: int) -> int:
    """
    Compute the CRT (q ** -1) % p value from RSA primes p and q.
    """
    return _modinv(q, p)


def rsa_crt_dmp1(private_exponent: int, p: int) -> int:
    """
    Compute the CRT private_exponent % (p - 1) value from the RSA
    private_exponent (d) and p.
    """
    return private_exponent % (p - 1)


def rsa_crt_dmq1(private_exponent: int, q: int) -> int:
    """
    Compute the CRT private_exponent % (q - 1) value from the RSA
    private_exponent (d) and q.
    """
    return private_exponent % (q - 1)


def rsa_recover_private_exponent(e: int, p: int, q: int) -> int:
    """
    Compute the RSA private_exponent (d) given the public exponent (e)
    and the RSA primes p and q.

    This uses the Carmichael totient function to generate the
    smallest possible working value of the private exponent.
    """
    # This lambda_n is the Carmichael totient function.
    # The original RSA paper uses the Euler totient function
    # here: phi_n = (p - 1) * (q - 1)
    # Either version of the private exponent will work, but the
    # one generated by the older formulation may be larger
    # than necessary. (lambda_n always divides phi_n)
    #
    # TODO: Replace with lcm(p - 1, q - 1) once the minimum
    # supported Python version is >= 3.9.
    lambda_n = (p - 1) * (q - 1) // gcd(p - 1, q - 1)
    return _modinv(e, lambda_n)


# Controls the number of iterations rsa_recover_prime_factors will perform
# to obtain the prime factors. Each iteration increments by 2 so the actual
# maximum attempts is half this number.
_MAX_RECOVERY_ATTEMPTS = 1000


def rsa_recover_prime_factors(n: int, e: int, d: int) -> tuple[int, int]:
    """
    Compute factors p and q from the private exponent d. We assume that n has
    no more than two factors. This function is adapted from code in PyCrypto.
    """
    # See 8.2.2(i) in Handbook of Applied Cryptography.
    ktot = d * e - 1
    # The quantity d*e-1 is a multiple of phi(n), even,
    # and can be represented as t*2^s.
    t = ktot
    while t % 2 == 0:
        t = t // 2
    # Cycle through all multiplicative inverses in Zn.
    # The algorithm is non-deterministic, but there is a 50% chance
    # any candidate a leads to successful factoring.
    # See "Digitalized Signatures and Public Key Functions as Intractable
    # as Factorization", M. Rabin, 1979
    spotted = False
    a = 2
    while not spotted and a < _MAX_RECOVERY_ATTEMPTS:
        k = t
        # Cycle through all values a^{t*2^i}=a^k
        while k < ktot:
            cand = pow(a, k, n)
            # Check if a^k is a non-trivial root of unity (mod n)
            if cand != 1 and cand != (n - 1) and pow(cand, 2, n) == 1:
                # We have found a number such that (cand-1)(cand+1)=0 (mod n).
                # Either of the terms divides n.
                p = gcd(cand + 1, n)
                spotted = True
                break
            k *= 2
        # This value was not any good... let's try another!
        a += 2
    if not spotted:
        raise ValueError("Unable to compute factors p and q from exponent d.")
    # Found !
    q, r = divmod(n, p)
    assert r == 0
    p, q = sorted((p, q), reverse=True)
    return (p, q)



import json
import logging
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

from order_api.dto import CustomProperty

from order_api import OrderService

from api.table import get_table_service
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from bustoken import TK_NCOUPON_APPLY_BENEFIT
from msgbus import FM_PARAM
from helper import (
    remove_accents,
    format_cpf_cnpj,
    format_phone,
)
from mw_helper import (
    show_check_list_box_dialog,
    show_filtered_list_box_dialog,
    show_loyalty_dialog,
    show_message_dialog,
    show_message_options_dialog,


)
from posot import OrderTaker
from sysactions import (
    action,
    get_current_order,
    get_model,
    get_posot,
    show_keyboard,
    show_messagebox,
    translate_message,
    StopAction,
)

from actions import mb_context
from actions.coupon.display_coupon import (
    get_benefits_from_order,
    remove_discounts,
)
from actions.customer import (
    get_customer_doc,
    get_customer_phone_number,
)
from actions.loyalty.model import LoyaltyButtonEnum
from application.interactor import LoyaltyInteractor
from utilconfigurations import get_ask_document_pinpad
from actions.loyalty.model import LoyaltyStopPaymentAction

from api.order import get_order_service

logger = logging.getLogger("PosActions")
title = "$INFORMATION"
loyalty_interactor = LoyaltyInteractor()
LOYALTY_ID = "LOYALTY_ID"
LOYALTY_CUSTOMER_INFO = "LOYALTY_CUSTOMER_INFO"
LOYALTY_CUSTOMER_BENEFITS = "LOYALTY_CUSTOMER_BENEFITS"
LOYALTY_COUPONS_TO_APPLY_ON_TOTAL = "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"
LOYALTY_PROPERTY_KEYS = [
    LOYALTY_ID,
    LOYALTY_CUSTOMER_INFO,
    LOYALTY_CUSTOMER_BENEFITS,
    LOYALTY_COUPONS_TO_APPLY_ON_TOTAL,
]


@action
def is_loyalty_component_available(
    _=None,     # type: str
):
    # type: (...) -> str

    loyalty_available = loyalty_interactor.is_loyalty_available()
    if loyalty_available is None:
        return ""

    return str(loyalty_available).lower()


@action
def open_loyalty(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> bool

    pos_id = int(pos_id)
    table_id = None if table_id in [None, "", "None"] else table_id
    order_id = None if order_id in [None, "", "None"] else int(order_id)

    while True:
        loyalty_properties = _get_loyalty_properties(
            pos_id=pos_id,
            table_id=table_id,
            order_id=order_id,
        )

        has_customer_info = False
        if loyalty_properties:
            has_customer_info = _has_customer_info(
                loyalty_properties=loyalty_properties,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        if has_customer_info:
            welcome_message = _get_welcome_message(
                loyalty_properties=loyalty_properties,
                order_id=order_id,
                table_id=table_id,
                table_seat=table_seat,
            )
            loyalty_id_button = "$REMOVE_LOYALTY"
            is_disabled = ""
        else:
            welcome_message, loyalty_id_button = loyalty_interactor.get_welcome_message_button_loyalty()
            is_disabled = '#DISABLED'

        buttons = "{}|$AVAILABLE_COUPONS{}|$APPLIED_COUPONS{}|$CLOSE".format(
            loyalty_id_button,
            is_disabled,
            is_disabled,
        )
        resp = show_loyalty_dialog(
            pos_id=str(pos_id),
            buttons=buttons,
            message=welcome_message,
        )

        if resp in (None, LoyaltyButtonEnum.CANCEL):
            break

        try:
            _execute_loyalty_option(
                pos_id=pos_id,
                response=resp,
                customer_info=has_customer_info,
                order_id=order_id,
                table_id=table_id,
                table_seat=table_seat,
            )
        except LoyaltyStopPaymentAction:
            pass

    return True


def _execute_loyalty_option(
    pos_id,         # type: int
    response,       # type: int
    customer_info,  # type: bool
    order_id,       # type: int
    table_id,       # type: str
    table_seat,     # type: str

):
    # type: (...) -> None

    if response == LoyaltyButtonEnum.UPDATE_CPF:
        if customer_info:
            _remove_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
        else:
            register_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        return

    if response == LoyaltyButtonEnum.AVAILABLE_COUPONS:
        if customer_info:
            _show_available_coupons(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        return

    if response == LoyaltyButtonEnum.APPLIED_COUPONS and customer_info:
        _show_applied_coupons(
            pos_id=pos_id,
            table_id=table_id,
            order_id=order_id,
            table_seat=table_seat,
        )


def register_loyalty_properties(
    pos_id,          # type: int
    table_id,        # type: Optional[str]
    order_id,        # type: Optional[int]
    table_seat,      # type: str
):
    # type: (...) -> None

    loyalty_id = _set_loyalty_properties(
        pos_id=pos_id,
        order_id=order_id,
        table_id=table_id,
        table_seat=table_seat,
        loyalty_type_id=loyalty_interactor.get_loyalty_id_type(),
    )

    if table_id:
        _update_table_picture(
            pos_id=pos_id,
            table_id=table_id,
        )

    if not loyalty_id or not order_id:
        return

    if not _is_order_totaled(pos_id=pos_id):
        return

    _apply_totaled_benefits(
        pos_id=pos_id,
        order_id=order_id,
    )


def _apply_totaled_benefits(
    pos_id,     # type: int
    order_id,   # type: int
):
    # type: (...) -> None

    order_service = get_order_service(pos_id=pos_id)
    order = order_service.get_order(order_id=order_id)
    to_apply_on_total = []
    benefits = []
    for custom_property in order.custom_properties:
        if custom_property.key == LOYALTY_COUPONS_TO_APPLY_ON_TOTAL:
            to_apply_on_total = json.loads(custom_property.value or "[]")
        if custom_property.key == LOYALTY_CUSTOMER_BENEFITS:
            benefits = json.loads(custom_property.value or "[]")

    benefits_to_apply = list(filter(lambda b: b["id"] in to_apply_on_total, benefits))

    data = json.dumps({
        "posId": pos_id,
        "originatorPosId": pos_id,
        "benefits": benefits_to_apply,
        "orderPicture": eTree.tostring(order.order_xml),
        "isManual": False,
    })

    mb_context.MB_EasySendMessage(
        dest_name="nCoupon",
        token=TK_NCOUPON_APPLY_BENEFIT,
        format=FM_PARAM,
        data=data,
    )


def _remove_loyalty_properties_qsr(
    pos_id,     # type: int
    order_id,   # type: int
):
    # type: (...) -> None

    properties_to_remove = LOYALTY_PROPERTY_KEYS + ["CUSTOMER_NAME", "CUSTOMER_DOC"]
    get_order_service(pos_id=pos_id).remove_order_custom_properties(
        keys=properties_to_remove,
        order_id=order_id,
    )


def _remove_loyalty_properties_ts(
    pos_id,                         # type: int
    table_id,                       # type: str
    table_seat,                     # type: str
):
    # type: (...) -> None

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
    )
    for prop in LOYALTY_PROPERTY_KEYS:
        if loyalty_properties.get(prop):
            loyalty_properties[prop].pop(table_seat, None)
            loyalty_properties[prop] = json.dumps(loyalty_properties[prop])

    if table_seat == "table":
        loyalty_properties["CUSTOMER_NAME"] = ""
        loyalty_properties["CUSTOMER_DOC"] = ""

    from msgbusboundary import MsgBusTableService as tableService

    tableService().set_table_custom_properties(
        pos_id=pos_id,
        properties=json.dumps(loyalty_properties),
        table_id=table_id,
    )
    _update_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )


def _update_table_picture(
    pos_id,     # type: int
    table_id,   # type: str
):
    # type: (...) -> None

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event

    table_picture = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )
    create_table_modified_event(
        pos_id=pos_id,
        table_picture=table_picture,
    )


def _get_loyalty_properties(
    pos_id,         # type: int
    table_id=None,  # type: Optional[str]
    order_id=None,  # type: Optional[int]
):
    # type: (...) -> Optional[Dict]

    if order_id:
        return _get_loyalty_properties_for_qsr(
            pos_id=pos_id,
            order_id=order_id,
        )

    if table_id:
        return _get_loyalty_properties_for_ts(
            pos_id=pos_id,
            table_id=table_id,
        )

    return None


def _get_loyalty_properties_for_qsr(
    pos_id,    # type: int
    order_id,  # type: int
):
    # type: (...) -> Optional[Dict]

    order_xml = get_order_service(pos_id=pos_id).get_order_picture(order_id=order_id)
    order_properties = order_xml.find(".//CustomOrderProperties")

    property_base = ".//OrderProperty[@key='{}']"
    customer_info = order_properties.find(property_base.format(LOYALTY_CUSTOMER_INFO))
    if customer_info is not None:
        customer_info = customer_info.get("value")

    benefits = order_properties.find(property_base.format(LOYALTY_CUSTOMER_BENEFITS))
    if benefits is not None:
        benefits = benefits.get("value")

    coupons_on_total = order_properties.find(property_base.format(LOYALTY_COUPONS_TO_APPLY_ON_TOTAL))
    if coupons_on_total is not None:
        coupons_on_total = coupons_on_total.get("value")

    return {
        LOYALTY_CUSTOMER_INFO: json.loads(customer_info) if customer_info else {},
        LOYALTY_CUSTOMER_BENEFITS: json.loads(benefits) if benefits else [],
        LOYALTY_COUPONS_TO_APPLY_ON_TOTAL: json.loads(coupons_on_total) if coupons_on_total else [],
    }


def _get_loyalty_properties_for_ts(
    pos_id,     # type: int
    table_id,   # type: str
):
    # type: (...) -> Optional[Dict]

    table_service = get_table_service(pos_id=pos_id)
    table_service_dto = table_service.get_service(table_id=table_id)
    table_properties = table_service_dto.custom_properties

    loyalty_ids = json.loads(table_properties.get(LOYALTY_ID, "{}"))

    customer_info = json.loads(table_properties.get(LOYALTY_CUSTOMER_INFO, "{}"))
    _parse_object_children(dictionary=customer_info)

    benefits = json.loads(table_properties.get(LOYALTY_CUSTOMER_BENEFITS, "{}"))
    _parse_object_children(dictionary=benefits)

    coupons_on_total = json.loads(table_properties.get(LOYALTY_COUPONS_TO_APPLY_ON_TOTAL, "{}"))
    _parse_object_children(dictionary=coupons_on_total)

    return {
        LOYALTY_ID: loyalty_ids,
        LOYALTY_CUSTOMER_INFO: customer_info,
        LOYALTY_CUSTOMER_BENEFITS: benefits,
        LOYALTY_COUPONS_TO_APPLY_ON_TOTAL: coupons_on_total,
    }


def _parse_object_children(
    dictionary,     # type: Dict
):
    # type: (...) -> None

    for key, value in dictionary.items():
        if isinstance(value, str):
            dictionary.update({key: json.loads(value)})


def _remove_loyalty_properties(
    pos_id,        # type: int
    table_id,      # type: Optional[str]
    order_id,      # type: Optional[int]
    table_seat,    # type: str
):
    # type: (...) -> None

    model = get_model(posid=pos_id)
    if _has_discounts_on_total(
        pos_id=pos_id,
        model=model,
    ):
        return

    if order_id:
        _remove_loyalty_properties_qsr(
            order_id=order_id,
            pos_id=pos_id,
        )

    if table_id:
        _remove_loyalty_properties_ts(
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
        )


def _set_loyalty_properties(
    pos_id,             # type: int
    loyalty_type_id,    # type: str
    table_id,           # type: Optional[str]
    order_id,           # type: Optional[int]
    table_seat,         # type: str
):
    # type: (...) -> Optional[str]

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)
    loyalty_id = get_user_input_by_type(
        pos_id=pos_id,
        pos_ot=pos_ot,
        loyalty_type_id=loyalty_type_id,
    )

    try:
        loyalty_customer_info = loyalty_interactor.set_loyalty_info(
            pos_id=pos_id,
            loyalty_id=loyalty_id,
            order_id=order_id,
            table_id=table_id,
            seat=table_seat,
        )
    except LoyaltyInteractor.LoyaltyPartnerError:
        logger.exception("Partner error")
        show_messagebox(
            posid=pos_id,
            message="$LOYALTY_PARTNER_ERROR",
        )
        return None

    if loyalty_customer_info is None:
        show_messagebox(
            posid=pos_id,
            message="$LOYALTY_INFORMATION_NOT_FOUND",
        )
        return None

    return loyalty_id


def _has_customer_info(
    loyalty_properties,     # type: Dict
    table_id,               # type: Optional[str]
    order_id,               # type: Optional[int]
    table_seat,             # type: str
):
    # type: (...) -> bool

    if order_id:
        customer_info = loyalty_properties.get(LOYALTY_CUSTOMER_INFO)
        return bool(customer_info)

    if table_id:
        customer_info = loyalty_properties.get(LOYALTY_CUSTOMER_INFO)
        customer_info = customer_info.get(table_seat, {})
        return bool(customer_info)

    return False


def _get_welcome_message(
    loyalty_properties,     # type: Dict
    order_id,               # type: Optional[int]
    table_id,               # type: Optional[str]
    table_seat,             # type: str
):
    # type: (...) -> unicode

    customer_info = loyalty_properties.get(LOYALTY_CUSTOMER_INFO, {})
    if table_id and not order_id:
        customer_info = customer_info.get(table_seat, {})

    customer_info = customer_info.get("customerInfo", {})
    customer = customer_info.get("customer", {})
    customer_name = customer.get("name") or "-"
    customer_doc = format_cpf_cnpj(customer.get("document")) or "-" if "document" in customer else "-"
    customer_phone = format_phone(customer.get("phone")) or "-" if "phone" in customer else "-"

    additional_info = customer_info.get("additionalInfo", {})
    welcome_message = additional_info.get("welcomeMessage", "")

    return "{}|{}|{}|{}|{}".format(
        "$WELCOME_CLIENT_LOYALTY_INFO",
        customer_name.encode("utf-8"),
        customer_doc,
        customer_phone,
        welcome_message.encode('utf-8'),
    )


def _show_available_coupons(
    pos_id,         # type: int
    table_id,       # type: str
    order_id,       # type: Optional[int]
    table_seat,     # type: str
):
    # type: (...) -> None

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )
    if not loyalty_properties:
        return

    customer_benefits, coupons_to_apply_on_total = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_seat=table_seat,
    )

    available_coupons = _get_available_coupons(
        pos_id=pos_id,
        customer_benefits=customer_benefits,
        coupons_to_apply_on_total=coupons_to_apply_on_total,
    )
    if not available_coupons:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$WITHOUT_ANY_LOYALTY_BENEFITS_TO_SHOW",
        )
        return

    options = [
        u"{} - {}".format(index + 1, coupon["coupon"]["name"]).encode("utf-8")
        for index, coupon in enumerate(available_coupons)
    ]
    selected_index = show_filtered_list_box_dialog(
        pos_id=str(pos_id),
        options=options,
        title="$SELECT_ONE_COUPON",
        mask="NOFILTER",
    )

    if selected_index is None:
        return

    if order_id:
        _save_selected_benefit_in_order(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            benefit_id=available_coupons[selected_index]["benefitId"],
            order_id=order_id,
        )
    else:
        _save_selected_benefit_in_table(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            benefit_id=available_coupons[selected_index]["benefitId"],
            table_id=table_id,
            table_seat=table_seat,
        )
    raise StopAction()


def _get_loyalty_coupons(
    loyalty_properties,             # type: Dict
    order_id,                       # type: Optional[int]
    table_seat,                     # type: str
    validate_from_source=False,     # type: Optional[bool]
):
    # type: (...) -> Tuple[Dict[str, Union[str, List, Dict]], List[str]]

    if order_id:
        customer_benefits = loyalty_properties.get(LOYALTY_CUSTOMER_BENEFITS, [])
        coupons_to_apply_on_total = loyalty_properties.get(LOYALTY_COUPONS_TO_APPLY_ON_TOTAL, [])
        if not coupons_to_apply_on_total and validate_from_source:
            coupons_to_apply_on_total = [b.get("id") for b in customer_benefits if b.get("automatic") == "ON_TOTAL"]

        return customer_benefits, coupons_to_apply_on_total

    customer_benefits = loyalty_properties.get(LOYALTY_CUSTOMER_BENEFITS, {})
    customer_benefits = customer_benefits.get(table_seat, [])
    coupons_to_apply_on_total = loyalty_properties.get(LOYALTY_COUPONS_TO_APPLY_ON_TOTAL, {})
    coupons_to_apply_on_total = coupons_to_apply_on_total.get(table_seat, [])

    return customer_benefits, coupons_to_apply_on_total


def _show_applied_coupons(
    pos_id,         # type: int
    table_id,       # type: Optional[str]
    order_id,       # type: Optional[int]
    table_seat,     # type: str
):
    # type: (...) -> None

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    customer_benefits, coupons_to_apply_on_total = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_seat=table_seat,
    )
    if not coupons_to_apply_on_total:
        show_messagebox(
            posid=pos_id,
            message="$ORDER_HAS_NO_APPLIED_DISCOUNT",
        )
        return

    coupon_options = _format_coupon_options(
        pos_id=pos_id,
        coupon_options=customer_benefits,
        coupons_to_apply_on_total=coupons_to_apply_on_total,
    )
    options = ["{0}".format(c["name"]) for c in coupon_options]
    if len(options) <= 1:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$WITHOUT_ANY_LOYALTY_BENEFITS_TO_SHOW",
        )
        return

    selected_indexes = show_check_list_box_dialog(
        pos_id=pos_id,
        options=options,
        title="$SELECT_ONE_COUPON",
        buttons="$REMOVE|$CANCEL",
        mask="NO_SALE_PANEL_IN_LOYALTY_BENEFITS",
    )
    if not len(selected_indexes):
        return

    loyalty_property = loyalty_properties[LOYALTY_COUPONS_TO_APPLY_ON_TOTAL]
    if not order_id:
        loyalty_property = loyalty_property[table_seat]

    indexes = [int(i) for i in selected_indexes if int(i)]
    for coupon_item in coupon_options:
        if coupon_item.get("index", 0) not in indexes:
            continue

        if order_id:
            _remove_loyalty_discount(
                pos_id=pos_id,
                order_id=order_id,
                coupon_item=coupon_item,
            )
        elif loyalty_property:
            loyalty_property.remove(coupon_item.get("id"))

    if not order_id:
        from msgbusboundary import MsgBusTableService as tableService

        tableService().set_table_custom_property(
            pos_id=pos_id,
            prop=LOYALTY_COUPONS_TO_APPLY_ON_TOTAL,
            prop_value=json.dumps(loyalty_properties[LOYALTY_COUPONS_TO_APPLY_ON_TOTAL]),
            table_id=table_id,
        )

    if _is_order_totaled(pos_id=pos_id):
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$ADD_ITEM_DISCOUNT_NOT_REMOVED",
        )


def _format_coupon_options(
    pos_id,                     # type: int
    coupon_options,             # type: Dict
    coupons_to_apply_on_total,  # type: List
):
    # type: (...) -> List

    model = get_model(posid=pos_id)
    all_coupons = translate_message(model, "ALL")
    applied_coupons = []
    benefit_index = 1

    order_benefits_ids = _get_order_benefits_ids_dict(model=model)
    is_order_totaled = _is_order_totaled(pos_id=pos_id)

    for coupon in coupon_options:
        benefit_id = coupon.get('id')
        benefit_key = order_benefits_ids[benefit_id] if benefit_id in order_benefits_ids.keys() else None
        add_item = _has_item_to_add(coupon=coupon)
        is_coupon_applied = benefit_id not in order_benefits_ids.keys()

        if (benefit_id not in coupons_to_apply_on_total) or (is_order_totaled and is_coupon_applied):
            continue

        disable = False
        description = (coupon.get("userDescription") or coupon.get("name") or coupon.get("id") or "").encode("UTF-8")
        name = "{}##{}##{}".format(description, 'null', disable)
        data = {
            "index": benefit_index,
            "id": benefit_id,
            "name": name,
            "benefit_key": benefit_key,
            "add_item": add_item
        }
        applied_coupons.append(data)
        benefit_index += 1

    name = "{}##{}##False".format(all_coupons, 'null')
    data = {"index": 0, "id": None, "name": name}
    applied_coupons.append(data)
    applied_coupons.sort(key=lambda x: (x["index"]))
    return applied_coupons


def _save_selected_benefit_in_table(
    loyalty_properties,     # type: Dict[str, Dict]
    pos_id,                 # type: int
    benefit_id,             # type: str
    table_id,               # type: Optional[str]
    table_seat,             # type: Optional[str]
):
    # type: (...) -> None

    coupons_to_apply_on_total = loyalty_properties[LOYALTY_COUPONS_TO_APPLY_ON_TOTAL]
    if table_seat in coupons_to_apply_on_total:
        coupons_to_apply_on_total[table_seat].append(benefit_id)
    else:
        coupons_to_apply_on_total[table_seat] = [benefit_id]

    from msgbusboundary import MsgBusTableService as tableService
    tableService().set_table_custom_property(
        pos_id=pos_id,
        prop=LOYALTY_COUPONS_TO_APPLY_ON_TOTAL,
        prop_value=json.dumps(coupons_to_apply_on_total),
        table_id=table_id,
    )

    show_messagebox(
        posid=pos_id,
        message="$APPLIED_DISCOUNTS_MUST_BE_APPLIED_WHEN_TOTAL",
    )


def _save_selected_benefit_in_order(
    loyalty_properties,     # type: Dict[str, Union[Dict, List]]
    pos_id,                 # type: int
    benefit_id,             # type: str
    order_id,               # type: int
):
    # type: (...) -> None

    coupons_to_apply_on_total = loyalty_properties[LOYALTY_COUPONS_TO_APPLY_ON_TOTAL]
    if coupons_to_apply_on_total is not None:
        coupons_to_apply_on_total.append(benefit_id)
    else:
        coupons_to_apply_on_total = [benefit_id]

    get_order_service(pos_id=pos_id).set_order_custom_property(
        order_id=order_id,
        custom_property=CustomProperty(
            key=LOYALTY_COUPONS_TO_APPLY_ON_TOTAL,
            value=json.dumps(coupons_to_apply_on_total),
        )
    )

    show_messagebox(
        posid=pos_id,
        message="$APPLIED_DISCOUNTS_MUST_BE_APPLIED_WHEN_TOTAL",
    )


def _is_tab(
    table_id,   # str
):
    # type: (...) -> bool

    return 'TAB' in table_id


def _get_available_coupons(
    pos_id,                     # type: int
    customer_benefits,          # type: List
    coupons_to_apply_on_total,  # type: List
):
    # type: (...) -> List[Dict]

    benefit_list = []
    model = get_model(posid=pos_id)
    order_benefits = _get_applied_benefit_on_order(model=model)
    order_benefits_ids = [benefit["benefit"]["id"] for benefit in order_benefits]
    automatic_benefit = remove_accents(translate_message(model, "AUTOMATIC_BENEFIT"))
    is_order_totaled = _is_order_totaled(pos_id=pos_id)

    for index, benefit in enumerate(customer_benefits):
        applied_benefit = _validate_applied_benefit(
            benefit=benefit,
            coupons_to_apply_on_total=coupons_to_apply_on_total,
            order_benefits_ids=order_benefits_ids,
            is_order_totaled=is_order_totaled,
        )

        if applied_benefit:
            continue

        name = str(benefit.get("userDescription", benefit.get("name", benefit.get("id", ""))).encode("utf-8"))
        is_automatic = benefit.get('automatic', None) not in [None, 'DISABLED']

        if name.startswith("$"):
            params = name.split("|", 1)[1] if "|" in name else ""
            message = name.split("|", 1)[0] if params else name
            name = translate_message(model, message[1:], *(params.split("|")))

        if is_order_totaled and is_automatic:
            name = name + " " + ("({0})".format(automatic_benefit))

        benefit_item = {
            "coupon": benefit,
            "name": name,
            "benefitId": benefit.get("id"),
        }
        benefit_list.append(benefit_item)

    return benefit_list


def _is_order_totaled(
    pos_id=None,     # type: int
    order=None,      # type: Element
):
    # type: (...) -> bool

    if not order:
        model = get_model(posid=pos_id)
        order = get_current_order(model=model)
    return order is not None and order.get("state") == "TOTALED"


def _get_order_benefits_ids_dict(
    model,     # type: Element
):
    # type: (...) -> Dict

    order_benefits = _get_applied_benefit_on_order(model=model)
    benefits_dict = {}
    for _, benefit in enumerate(order_benefits):
        benefits_dict.update({benefit["benefit"]["id"]: benefit["benefit_key"]})
    return benefits_dict


def _has_item_to_add(
    coupon,     # type: Dict
):
    # type: (...) -> Optional[bool]

    coupon_actions = coupon.get("actions")
    for coupon_action in coupon_actions.values():
        if coupon_action.get("itemsToAdd"):
            return True


def _remove_loyalty_discount(
    pos_id,                 # type: int
    order_id,               # type: int
    coupon_item,            # type: Dict[str, Union[int, str]]
):
    # type: (...) -> None

    order_service = get_order_service(pos_id=pos_id)
    order = order_service.get_order_picture(order_id=order_id)
    coupon_key = _get_applied_benefit_key(
        order=order,
        coupon=coupon_item,
    )
    if coupon_key:
        remove_discounts(
            pos_id=pos_id,
            coupon_ids=[coupon_key],
        )
    else:
        _remove_from_loyalty_on_total_qsr(
            order_service=order_service,
            order=order,
            benefit_id=coupon_item.get("id"),
        )


def _get_applied_benefit_key(
    order,      # type: eTree.Element
    coupon,     # type: str
):
    # type: (...) -> Optional[str]

    if coupon.get("benefit_key"):
        return coupon["benefit_key"]

    benefit_list_custom_property = order.find(".//OrderProperty[@key='BENEFIT_LIST']")
    if benefit_list_custom_property is None:
        return None

    benefit_list_json = json.loads(benefit_list_custom_property.get("value"))
    benefit_json = benefit_list_json.get("benefitList")
    benefit_list = json.loads(benefit_json)
    for benefit in benefit_list:
        if benefit["benefit"]["id"] == coupon.get("id"):
            return benefit["benefit_key"]

    return None


def _remove_from_loyalty_on_total_qsr(
    order_service,  # type: OrderService
    order,          # type: eTree.Element
    benefit_id,     # type: str
):
    # type: (...) -> None

    prop = order.find(".//OrderProperty[@key='{}']".format(LOYALTY_COUPONS_TO_APPLY_ON_TOTAL))
    if prop is None:
        return

    loyalty_on_total = json.loads(prop.get("value") or "[]")
    new_loyalty_on_total = []
    for current_benefit_id in loyalty_on_total:
        if current_benefit_id != benefit_id:
            new_loyalty_on_total.append(current_benefit_id)

    order_service.set_order_custom_property(custom_property=CustomProperty(
        key=LOYALTY_COUPONS_TO_APPLY_ON_TOTAL,
        value=json.dumps(new_loyalty_on_total),
    ))


def _validate_applied_benefit(
    benefit,                    # type: Dict
    coupons_to_apply_on_total,  # type: List
    order_benefits_ids,         # type: List
    is_order_totaled,           # type: bool
):
    # type: (...) -> bool

    applied_benefit = benefit.get("id") in coupons_to_apply_on_total
    benefit_applied_in_order = benefit.get("id") in order_benefits_ids
    if is_order_totaled:
        applied_benefit = applied_benefit and benefit_applied_in_order

    return applied_benefit


def _get_applied_benefit_on_order(
    model,     # type: Element
):
    # type: (...) -> Dict

    pos_ot = get_posot(model=model)
    order = get_current_order(model=model)
    order_benefits = get_benefits_from_order(
        pos_ot=pos_ot,
        order=order,
    )
    return order_benefits


def _has_discounts_on_total(
    pos_id,     # type: int
    model,      # type: Element
):
    # type: (...) -> bool

    order_benefits = _get_applied_benefit_on_order(model=model)
    is_order_totaled = _is_order_totaled(pos_id=pos_id)
    if order_benefits and is_order_totaled:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$CANNOT_REMOVE_LOYALTY_ON_TOTALED_SCREEN",
        )
        return True


def get_user_input_by_type(
    pos_id,             # type: int
    pos_ot,             # type: OrderTaker
    loyalty_type_id,    # type: str
):
    # type: (...) -> str

    if loyalty_type_id == "DOCUMENT":
        return _get_user_document_input(
            pos_id=pos_id,
            pos_ot=pos_ot,
            default_doc="",
            stop_action=False,
        )

    if loyalty_type_id == "PHONE":
        return _get_user_phone_input(
            pos_id=pos_id,
            stop_action=False,
        )

    if loyalty_type_id == "OTHER":
        return _get_user_other_input(pos_id=pos_id)


def _get_user_phone_input(
    pos_id,         # type: int
    stop_action,    # type: bool
):
    # type: (...) -> str

    loyalty_id = get_customer_phone_number(
        pos_id=pos_id,
        stop_action=stop_action,
    )

    if loyalty_id is None:
        raise LoyaltyStopPaymentAction()

    return loyalty_id


def _get_user_document_input(
    pos_id,         # type: int
    pos_ot,         # type: OrderTaker
    default_doc,    # type: str
    stop_action,    # type: bool
):
    # type: (...) -> str

    document_pin_pad = get_ask_document_pinpad(pos_id=pos_id)
    loyalty_id = get_customer_doc(
        pos_id=pos_id,
        pos_ot=pos_ot,
        default_doc=default_doc,
        ask_document_pinpad=document_pin_pad,
        stop_action=stop_action,
    )
    if loyalty_id is None:
        raise LoyaltyStopPaymentAction()

    return loyalty_id


def _get_user_other_input(
    pos_id,     # type: int
):
    # type: (...) -> str

    loyalty_id = show_keyboard(
        posid=pos_id,
        message="$ENTER_LOYALTY_ID",
        buttons="$OK|$NOT_INFORM|$PREVIOUS",
        defvalue="",
        title="",
    )
    if loyalty_id.strip() in ("", "NOT_INFORM"):
        return ""

    if loyalty_id is None:
        raise LoyaltyStopPaymentAction()

    return loyalty_id.translate(None, '-+[]/;.|<>:?{}=_()*&^%$#@!,').upper()


def get_table_loyalty_customer_attribute(
    pos_id,                 # type: str
    table_id,               # type: str
    attribute="document",   # type: Optional[str]
):
    # type: (...) -> str

    if table_id is None:
        return ''

    from msgbusboundary import MsgBusTableService as tableService

    loyalty_info = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop=LOYALTY_CUSTOMER_INFO,
        table_id=table_id,
    )
    if loyalty_info:
        loyalty_info = json.loads(loyalty_info.get(LOYALTY_CUSTOMER_INFO))
        customer = loyalty_info.get("table", {}).get("customerInfo", {}).get("customer", {})
        return customer.get(attribute, "")
    return ""


def get_order_loyalty_customer_attribute(
    pos_id,                 # type: str
    order_picture,          # type: Element
    attribute="document",   # type: Optional[str]
):
    # type: (...) -> str

    loyalty_info = order_picture.find(".//OrderProperty[@key='{}']".format(LOYALTY_CUSTOMER_INFO))
    table_id = order_picture.find(".//OrderProperty[@key='TABLE_ID']")
    if loyalty_info is not None:
        return json.loads(loyalty_info.get("value")).get("customerInfo", {}).get("customer", {}).get(attribute, "")
    elif table_id is not None:
        return get_table_loyalty_customer_attribute(
            pos_id=pos_id,
            table_id=table_id.get("value"),
            attribute=attribute,
        )
    return ""


def get_document_from_loyalty_or_customer_response(
    pos_id,                     # type: str
    pos_ot,                     # type: OrderTaker
    ask_document_pinpad=False,  # type: bool
    document=None,              # type: Optional[str]
    stop_action=True,           # type: Optional[bool]
):
    # type: (...) -> Optional[str]

    if document is not None:
        return None

    order_picture = eTree.XML(pos_ot.orderPicture(posid=pos_id))
    property_document = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='CUSTOMER_DOC']")
    if property_document is not None and property_document.get("value"):
        return None

    is_valid_document = get_order_loyalty_customer_attribute(
        pos_id=pos_id,
        order_picture=order_picture,
        attribute="isValidDocument",
    )

    if is_valid_document:
        loyalty_document = get_order_loyalty_customer_attribute(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        buttons = "$CONFIRM|$OTHER|$WITHOUT_DOCUMENT"
        message = "$CONFIRM_CPF_CNPJ|{0}".format(format_cpf_cnpj(value=loyalty_document))
        resp = show_message_options_dialog(
            pos_id=pos_id,
            buttons=buttons,
            title="$CONFIRMATION",
            message=message,
        )
        if resp == 2 or resp is None:
            return ""
        if resp == 0:
            return loyalty_document

    return get_customer_doc(
        pos_id=pos_id,
        pos_ot=pos_ot,
        ask_document_pinpad=ask_document_pinpad,
        stop_action=stop_action,
    ) or ""



import json
import logging

from typing import Dict

from requests import Response

from typing import Union


class NotFound(Exception):
    pass


class BackupAlertsDatabaseNoSourceException(Exception):
    pass


class RestoreAlertsDatabaseNoSourceException(Exception):
    pass


class NoResponseError(Exception):

    def __init__(
        self,
        logger,     # logging.Logger
        response,   # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "No response. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(NoResponseError, self).__init__(content)


class ForbiddenError(Exception):

    def __init__(
        self,
        logger,    # logging.Logger
        response,  # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "Forbidden error. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(ForbiddenError, self).__init__(content)


class NotFoundError(Exception):

    def __init__(
        self,
        logger,    # logging.Logger
        response,  # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "Not found. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(NotFoundError, self).__init__(content)


class NotAvailableError(Exception):

    def __init__(
        self,
        logger,    # logging.Logger
        response,  # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "Code not available. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(NotAvailableError, self).__init__(content)


class InUseError(Exception):

    def __init__(
        self,
        logger,    # logging.Logger
        response,  # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "Code in use on another terminal. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(InUseError, self).__init__(content)


class ProcessorError(Exception):

    def __init__(
        self,
        logger,    # logging.Logger
        response,  # Response
    ):
        # type: (...) -> None

        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)

        if self.logger.isEnabledFor(logging.ERROR):
            response_str = json_serialize(response)
            self.logger.error(
                "Processing error. StatusCode: %r; Content: %r; FullResponse: %r",
                error_status_code,
                content,
                response_str,
            )

        super(ProcessorError, self).__init__(content)


def _get_response_content(
    response,  # type: Response
):
    # type: (...) -> str

    try:
        content = json.loads(response.content).get("detail").encode("utf8")
        return content
    except (AttributeError, ValueError):
        return ""


def json_serialize(
    obj,  # type: Union[str, Dict]
):
    # type: (...) -> str

    try:
        return json.dumps(obj, default=lambda o: getattr(o, '__dict__', str(o)))
    except ValueError:  # noqa
        return obj




class CustomOrderProperty(object):
    def __init__(self, key, value):
        self.__key = key
        self.__value = value

    @property
    def key(self):
        # type: () -> str
        return self.__key

    @property
    def value(self):
        # type: () -> str
        return self.__value
